from __future__ import annotations
import abc
import json
import email
import h11
import socket
import time
import os
import trio
import typing as t
from rsyscall.trio_test_case import TrioTestCase
from rsyscall.thread import Thread, ChildThread
from rsyscall.handle import FileDescriptor, Path, WrittenPointer, Pointer
from rsyscall.command import Command
from rsyscall.memory.ram import RAM
from rsyscall.monitor import AsyncChildProcess
from dataclasses import dataclass
import rsyscall.tasks.local as local
from rsyscall.mktemp import update_symlink
from rsyscall.nix import local_store
from rsyscall.tasks.stub import StubServer

from rsyscall.inotify_watch import Inotify

from rsyscall.sched import CLONE
from rsyscall.netinet.in_ import SockaddrIn
from rsyscall.sys.inotify import IN
from rsyscall.sys.socket import AF, SOCK, Sockbuf
from rsyscall.fcntl import O
from rsyscall.sys.un import SockaddrUn

class JobsetInput:
    @abc.abstractmethod
    def type(self) -> str: ...
    @abc.abstractmethod
    def value(self) -> str: ...
    def dict(self) -> t.Dict[str, str]:
        return {"type": self.type(), "value": self.value()}

class JobsetIncludeInput(JobsetInput):
    """"

    The subset of jobset inputs which get put on the Nix include path,
    and therefore are valid candidates for finding the jobset input
    expression.

    """    
    pass

@dataclass
class PathInput(JobsetIncludeInput):
    path: Path
    def type(self) -> str:
        return "path"
    def value(self) -> str:
        return os.fsdecode(self.path)

@dataclass
class StringInput(JobsetInput):
    string: str
    def type(self) -> str:
        return "string"
    def value(self) -> str:
        return self.string

def lookup_alist(alist: t.List[t.Tuple[bytes, bytes]], find_key: bytes) -> t.Optional[bytes]:
    for key, value in alist:
        if key == find_key:
            return value
    return None

class HTTPClient:
    """Should an HTTP client object be called HTTPServer or HTTPClient?

    It represents a server locally - so perhaps it should be server.

    But most people call it client - so perhaps it should be client.

    I'll conform with the masses...

    """
    def __init__(self, read: t.Callable[[], t.Awaitable[bytes]],
                 write: t.Callable[[bytes], t.Awaitable[None]],
                 headers: t.List[t.Tuple[str, str]]) -> None:
        self.read = read
        self.write = write
        self.headers = headers
        self.connection = h11.Connection(our_role=h11.CLIENT)
        self.cookie: t.Optional[bytes] = None

    @staticmethod
    async def connect_unix(thread: Thread, addr: WrittenPointer[SockaddrUn]) -> HTTPClient:
        sock = await thread.make_afd(await thread.socket(AF.UNIX, SOCK.STREAM|SOCK.NONBLOCK))
        await sock.connect(addr)
        return HTTPClient(sock.read_some_bytes, sock.write_all_bytes, [
            ("Host", "localhost"),
            ("Accept", "application/json"),
            ("Content-Type", "application/json"),
        ])

    @staticmethod
    async def connect_inet(thread: Thread, addr: SockaddrIn) -> HTTPClient:
        sock = await thread.make_afd(await thread.socket(AF.INET, SOCK.STREAM|SOCK.NONBLOCK))
        await sock.connect(await thread.ram.ptr(addr))
        return HTTPClient(sock.read_some_bytes, sock.write_all_bytes, [
            ("Host", "localhost"),
            ("Accept", "application/json"),
            ("Content-Type", "application/json"),
        ])

    async def next_event(self) -> t.Any:
        while True:
            event = self.connection.next_event()
            if event is h11.NEED_DATA:
                data = await self.read()
                self.connection.receive_data(data)
                continue
            return event

    def get_headers(self) -> t.List[t.Tuple[str, str]]:
        if self.cookie is None:
            return self.headers
        else:
            return [*self.headers, ("Cookie", self.cookie.decode())]

    async def post(self, target: str, body: bytes) -> bytes:
        # send request
        data = self.connection.send(h11.Request(
            method="POST", target=target,
            headers=[*self.get_headers(), ("Content-Length", str(len(body)))]))
        data += self.connection.send(h11.Data(data=body))
        data += self.connection.send(h11.EndOfMessage())
        await self.write(data)
        # get response
        response = await self.next_event()
        data     = await self.next_event()
        eom      = await self.next_event()
        print(response); print(data); print(eom)
        set_cookie = lookup_alist(response.headers, b"set-cookie")
        if set_cookie is not None:
            self.cookie = set_cookie
        if response.status_code >= 300:
            raise Exception("error posting", data.data)
        self.connection.start_next_cycle()
        return data.data

    async def put(self, target: str, body: bytes) -> bytes:
        request = h11.Request(
            method="PUT", target=target,
            headers=[*self.get_headers(), ("Content-Length", str(len(body)))])
        print("PUT request", request)
        data = self.connection.send(request)
        data += self.connection.send(h11.Data(data=body))
        data += self.connection.send(h11.EndOfMessage())
        await self.write(data)
        response = await self.next_event()
        data     = await self.next_event()
        eom      = await self.next_event()
        print(response, data, eom)
        if response.status_code >= 300:
            raise Exception("error posting", data.data)
        self.connection.start_next_cycle()
        return data.data

    async def get(self, target: str) -> bytes:
        request = h11.Request(method="GET", target=target, headers=self.get_headers())
        print("request", request)
        data = self.connection.send(request)
        data += self.connection.send(h11.EndOfMessage())
        await self.write(data)
        response = await self.next_event()
        data = await self.next_event()
        eom = await self.next_event()
        print(response, data, eom)
        if response.status_code >= 300:
            raise Exception("error getting", data.data)
        self.connection.start_next_cycle()
        return data.data
        

@dataclass
class Postgres:
    sockdir: Path
    thread: Thread
    createuser_cmd: Command
    createdb_cmd: Command

    async def createuser(self, name: str) -> None:
        await self.thread.run(self.createuser_cmd.args("--host", self.sockdir, "--no-password", name))

    async def createdb(self, name: str, owner: str) -> None:
        await self.thread.run(self.createdb_cmd.args("--host", self.sockdir, "--owner", owner, name))

async def read_completely(ram: RAM, fd: FileDescriptor) -> bytes:
    data = b''
    while True:
        buf = await ram.malloc(bytes, 4096)
        valid, rest = await fd.pread(buf, offset=len(data))
        if valid.size() == 0:
            return data
        data += await valid.read()

async def start_postgres(nursery, thread: Thread, path: Path) -> Postgres:
    initdb = await thread.environ.which("initdb")
    postgres = await thread.environ.which("postgres")
    createuser = await thread.environ.which("createuser")
    createdb = await thread.environ.which("createdb")

    data = path/"data"
    await thread.run(initdb.args("--pgdata", data, "--nosync", "--no-locale", "--auth=trust"))
    sockdir = path/"sock"
    await thread.mkdir(sockdir)
    config = {
        # connection
        "listen_addresses": "''",
        "unix_socket_directories": f"'{os.fsdecode(sockdir)}'",
        # performance
        "fsync": "off",
        "synchronous_commit": "off",
        "full_page_writes": "off",
    }
    await thread.spit(data/"postgresql.auto.conf", "\n".join([f"{key} = {value}" for key, value in config.items()]))
    inty = await Inotify.make(thread)
    watch = await inty.add(data, IN.CLOSE_WRITE)

    await nursery.start(thread.run, postgres.args('-D', data))
    # pg_ctl uses the pid file to determine when postgres is up, so we do the same.
    pid_file_name = "postmaster.pid"
    pid_file = None
    name = await thread.ram.ptr(data/pid_file_name)
    while True:
        await watch.wait_until_event(IN.CLOSE_WRITE, pid_file_name)
        if pid_file is None:
            pid_file = await thread.task.open(name, O.RDONLY)
        pid_file_data = await read_completely(thread.ram, pid_file)
        try:
            # the postmaster status is on line 7
            # would be nice to get that from LOCK_FILE_LINE_PM_STATUS in pidfile.h
            pm_status = pid_file_data.split(b'\n')[7]
        except IndexError:
            pm_status = b""
        if b"ready" in pm_status:
            break
    await inty.close()
    return Postgres(sockdir, thread, createuser, createdb)

class NginxChild:
    # can support methods for reloading configuration, etc
    def __init__(self, child: AsyncChildProcess) -> None:
        self.child = child

async def exec_nginx(thread: ChildThread, nginx: Command,
                     path: Path, config: FileDescriptor,
                     listen_fds: t.List[FileDescriptor]) -> AsyncChildProcess:
    nginx_fds = [fd.maybe_copy(thread.task) for fd in listen_fds]
    config_fd = config.maybe_copy(thread.task)
    await thread.unshare_files()
    if nginx_fds:
        nginx_var = ";".join([str(await fd.as_argument()) for fd in nginx_fds]) + ';'
    else:
        nginx_var = ""
    await thread.mkdir(path/"logs")
    child = await thread.exec(
        nginx.env(NGINX=nginx_var).args("-p", path, "-c", f"/proc/self/fd/{await config_fd.as_argument()}"))
    return child

async def start_fresh_nginx(
        nursery, parent: Thread, path: Path, proxy_addr: SockaddrUn
) -> t.Tuple[SockaddrIn, NginxChild]:
    nginx = await parent.environ.which("nginx")
    thread = await parent.clone(CLONE.NEWUSER|CLONE.NEWPID)
    sock = await thread.task.socket(AF.INET, SOCK.STREAM)
    zero_addr = await thread.ram.ptr(SockaddrIn(0, 0x7F_00_00_01))
    await sock.bind(zero_addr)
    addr = await (await (await sock.getsockname(
        await thread.ram.ptr(Sockbuf(zero_addr)))).read()).buf.read()
    config = b"""
error_log stderr error;
daemon off;
events {}
http {
  access_log /proc/self/fd/1 combined;
  server {
    listen localhost:%d;
    location / {
        proxy_pass http://unix:%s;
    }
  }
}
""" % (addr.port, proxy_addr.path)
    await sock.listen(10)
    config_fd = await thread.task.open(await thread.ram.ptr(path/"nginx.conf"),
                                       O.RDWR|O.CREAT)
    remaining: Pointer = await thread.ram.ptr(config)
    while remaining.size() > 0:
        _, remaining = await config_fd.write(remaining)
    child = await exec_nginx(thread, nginx, path, config_fd, [sock])
    nursery.start_soon(child.check)
    return addr, NginxChild(child)

async def start_simple_nginx(nursery, parent: Thread, path: Path, sockpath: Path) -> NginxChild:
    nginx = await parent.environ.which("nginx")
    thread = await parent.clone(CLONE.NEWUSER|CLONE.NEWPID)
    config = b"""
error_log stderr error;
daemon off;
events {}
http {
  access_log /proc/self/fd/1 combined;
  server {
    listen localhost:3000;
    location / {
        proxy_pass http://unix:%s;
    }
  }
}
""" % os.fsencode(sockpath)
    sock = await thread.task.socket(AF.INET, SOCK.STREAM)
    await sock.bind(await thread.ram.ptr(SockaddrIn(3000, 0x7F_00_00_01)))
    await sock.listen(10)
    config_fd = await thread.task.open(await thread.ram.ptr(path/"nginx.conf"),
                                       O.RDWR|O.CREAT)
    remaining: Pointer = await thread.ram.ptr(config)
    while remaining.size() > 0:
        _, remaining = await config_fd.write(remaining)
    child = await exec_nginx(thread, nginx, path, config_fd, [sock])
    nursery.start_soon(child.check)
    return NginxChild(child)

@dataclass
class Hydra:
    addr: WrittenPointer[SockaddrUn]

class Store:
    @abc.abstractmethod
    def uri(self) -> str: ...

def build_url(root: str, **params: t.Optional[Path]) -> str:
    result = root
    first = True
    for key, value in params.items():
        if value is None:
            continue
        elif isinstance(value, Path):
            string = os.fsdecode(value)
        else:
            string = value
        if first:
            result += f"?{key}={string}"
            first = False
        else:
            result += f"&{key}={string}"
    return result

class LocalStore(Store):
    # Hydra requires the state dir to exist in the local filesystem so it can make gc roots
    # TODO let's robustify our handling of all this
    # hmmmmmm
    # so Hydra seems hardcoded to store logs in a subdirectory of hydraData.
    # actually, how does it even use this logDir?
    # okay, so... it streams the logs over to the central hydra host.
    # why not directly upload them?
    # why not just have the store on that host upload them... hmm....
    # so I see, we upload the dependencies to the host, do the build, then
    # pull the build results back I guess
    # ok so it's just that it wants to create gc roots
    # h m m
    @abc.abstractmethod
    def state_dir(self) -> Path: ...

@dataclass
class DirectStore(LocalStore):
    store: t.Optional[Path] = None
    state: t.Optional[Path] = None

    def state_dir(self) -> Path:
        if self.state is None:
            raise Exception("hmm")
        else:
            return self.state

    def uri(self) -> str:
        return build_url("local", store=self.store, state=self.state)

@dataclass
class DaemonStore(Store):
    def uri(self) -> str:
        return "daemon"

async def start_hydra(nursery, thread: Thread, path: Path, dbi: str, store: LocalStore) -> Hydra:
    # maybe have a version of this function which uses cached path locations?
    # or better yet, compiled in locations?
    hydra_init = await thread.environ.which("hydra-init")
    hydra_create_user = await thread.environ.which("hydra-create-user")
    hydra_server = await thread.environ.which("hydra-server")
    hydra_evaluator = await thread.environ.which("hydra-evaluator")
    hydra_queue_runner = await thread.environ.which("hydra-queue-runner")

    await thread.run(hydra_init.env(HYDRA_DBI=dbi, HYDRA_DATA=path))
    await thread.run(hydra_create_user.args(
        "sbaugh",
        "--full-name", "Spencer Baugh",
        "--email-address", "sbaugh@localhost",
        "--password", "foobar",
        "--role", "admin",
    ).env(HYDRA_DBI=dbi, HYDRA_DATA=path))

    # create config files
    config = {"email_notification": "1"}
    config_path = await thread.spit(path/"hydra.conf", "\n".join([f"{key} = {value}" for key, value in config.items()]))
    hydra_env: t.Mapping[str, t.Union[str, bytes, os.PathLike]] = {
        'HYDRA_DBI': dbi,
        'HYDRA_DATA': path,
        'HYDRA_CONFIG': config_path,
        'NIX_REMOTE': store.uri(),
        'NIX_STATE_DIR': store.state_dir(),
    }
    # start server
    server_thread = await thread.clone()
    await server_thread.unshare_files()
    sock = await server_thread.task.socket(AF.UNIX, SOCK.STREAM)
    addr = await server_thread.ram.ptr(
        await SockaddrUn.from_path(server_thread, path/"hydra_server.sock"))
    await sock.bind(addr)
    await sock.listen(10)
    ssport = os.fsdecode(addr.value.path)+"="+str(int(sock.near))
    server_child = await server_thread.exec(hydra_server.env(
        **hydra_env, SERVER_STARTER_PORT=ssport))
    nursery.start_soon(server_child.check)
    # start evaluator, queue runner
    await nursery.start(thread.run, hydra_evaluator.env(**hydra_env))
    await nursery.start(thread.run, hydra_queue_runner.env(**hydra_env))
    return Hydra(addr)

class Jobset:
    def __init__(self, project: Project, identifier: str) -> None:
        self.project = project
        self.identifier = identifier

class Project:
    def __init__(self, client: HydraClient, identifier: str) -> None:
        self.client = client
        self.identifier = identifier

    async def make_jobset(
            self, identifier: str,
            nixexprinput: JobsetIncludeInput, nixexprpath: str,
            inputs: t.Dict[str, JobsetInput],
            description: str=None,
            emailoverride: str="",
    ) -> Jobset:
        if description is None:
            description = identifier
        for name, input in inputs.items():
            if input is nixexprinput:
                nixexprinput_name = name
                break
        else:
            nixexprinput_name = "nixexprinput"
            if nixexprinput_name in inputs:
                raise Exception(f"tried to default to name '{nixexprinput_name}' for the Nix expr input, "
                                "but you already used that name in your input dictionary. "
                                "Either explicitly include the Nix expr input in your input dictionary under any name, "
                                f"or don't use the name '{nixexprinput_name}'.")
            inputs = {**inputs, nixexprinput_name: nixexprinput}
        await self.client.http.put(f'/jobset/{self.identifier}/{identifier}', json.dumps({
            "identifier": identifier,
            "description": description,
            "checkinterval": "60",
            "enabled": "1",
            "visible": "1",
            "keepnr": "1",
            "nixexprinput": nixexprinput_name,
            "nixexprpath": nixexprpath,
            "enableemail": "1" if emailoverride else "0",
            "emailoverride": emailoverride,
            "inputs": {name:input.dict() for name, input in inputs.items()},
        }).encode())
        return Jobset(self, identifier)

class HydraClient:
    @staticmethod
    async def login(http: HTTPClient) -> 'HydraClient':
        await http.post("/login", json.dumps({'username': "sbaugh", 'password': "foobar"}).encode())
        return HydraClient(http)

    def __init__(self, http: HTTPClient) -> None:
        self.http = http

    async def make_project(self, identifier: str, displayname: str=None) -> Project:
        if displayname is None:
            displayname = identifier
        await self.http.put(f'/project/{identifier}', json.dumps({
            'identifier': identifier, 'displayname': displayname, 'enabled': '1', 'visible': '1',
        }).encode())
        return Project(self, identifier)

class TestHydra(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thread = local.thread
        self.tmpdir = await self.thread.mkdtemp("test_hydra")
        self.path = self.tmpdir.path
        await update_symlink(self.thread, await self.thread.ram.ptr(self.tmpdir.parent/"test_hydra.current"), self.path)

        self.postgres = await start_postgres(self.nursery, self.thread, await self.thread.mkdir(self.path/"postgres"))
        await self.postgres.createuser("hydra")
        await self.postgres.createdb("hydra", owner="hydra")

        stubbin = await self.thread.mkdir(self.path/"stubbin")
        self.sendmail_stub = await StubServer.make(self.thread, local_store, stubbin, "sendmail")
        self.thread.environ['PATH'] = os.fsdecode(stubbin) + ':' + self.thread.environ['PATH']
        # start server
        # TODO I suppose this pidns thread is just going to be GC'd away... gotta make sure that works fine.
        pidns_thread = await self.thread.clone(CLONE.NEWUSER|CLONE.NEWPID)
        self.hydra = await start_hydra(
            self.nursery, pidns_thread, await self.thread.mkdir(self.path/"hydra"),
            "dbi:Pg:dbname=hydra;host=" + os.fsdecode(self.postgres.sockdir) + ";user=hydra;",
            DirectStore(store=self.path/"nix"/"store", state=self.path/"nix"/"state"),
        )

    async def asyncTearDown(self) -> None:
        await self.tmpdir.cleanup()

    async def create_and_validate_job(self, client: HydraClient) -> None:
        project = await client.make_project('neato', "A neat project")
        job_name = "jobbymcjobface"
        jobset_dir = await self.thread.mkdir(self.path/"jobset")
        jobset_path = "trivial.nix"
        await self.thread.spit(jobset_dir/jobset_path, """{ string }:
{ """ + job_name + """ = builtins.derivation {
    name = "trivial";
    system = "x86_64-linux";
    builder = "/bin/sh";
    args = ["-c" "echo ${string} > $out; exit 0"];
  };
}""")
        email_address = "foo@bar"
        jobset = await project.make_jobset('trivial', PathInput(jobset_dir), jobset_path, {
            "string": StringInput("Hello World"),
        }, description="Some trivial jobset", emailoverride=email_address)

        args, sendmail_task = await self.sendmail_stub.accept()
        print("sendmail args", args)
        data = await sendmail_task.read_to_eof(sendmail_task.stdin)
        message = email.message_from_bytes(data)
        self.assertEqual(email_address,  message['To'])
        self.assertIn("Success", str(message['Subject']))
        self.assertEqual(project.identifier, message['X-Hydra-Project'])
        self.assertEqual(jobset.identifier,  message['X-Hydra-Jobset'])
        self.assertEqual(job_name, message['X-Hydra-Job'])

    async def test_hydra(self) -> None:
        client = await HydraClient.login(await HTTPClient.connect_unix(self.thread, self.hydra.addr))
        await self.create_and_validate_job(client)

    async def test_proxy(self) -> None:
        addr, _ = await start_fresh_nginx(self.nursery, self.thread, await self.thread.mkdir(self.path/"nginx"),
                                          self.hydra.addr.value)
        client = await HydraClient.login(await HTTPClient.connect_inet(self.thread, addr))
        await self.create_and_validate_job(client)

if __name__ == "__main__":
    import unittest
    unittest.main()



"""Documentation for the design of the system call API in rsyscall

## manpages are useful
rsyscall provides a type-safe, low-level, broad interface to a variety of Linux functionality.
This includes basic system calls,
like the socket API or read and write.
It also includes less common interfaces, such as inotify, rtnetlink, timerfd, eventfd, and others.

The Linux kernel manpages are the primary documentation for development against Linux.
This is true for rsyscall and Python just as it is true for C.

The manpages are therefore the primary documentation for programming with rsyscall.
All system calls in rsyscall are exposed in such a way
that a mechanical reading of the manpage for the system call
will provide sufficient information to use the system call in rsyscall.

We will discuss this translation through several examples,
and at the end we will have a small set of simple rules,
which one may apply to program with any supported system call in rsyscall.

## System calls and constants
The manpage for any Linux system call begins with a `SYNOPSIS` section,
which shows the prototype for the glibc wrapper for the system call,
and shows the headers which should be included to use that system call.

For example, [socket(2)](https://www.man7.org/linux/man-pages/man2/socket.2.html) begins:

```c
#include <sys/types.h>          /* See NOTES */
#include <sys/socket.h>

int socket(int domain, int type, int protocol);
```

The declaration of `socket` and associated constants is provided by the `sys/socket.h` header.

Sometimes, as with `socket(2)`,
additional headers (such as `sys/types.h`) will be present in the `SYNOPSIS` for backwards compatibility;
the last header listed is always the header containing the declaration.

As described in [socket(2)](https://www.man7.org/linux/man-pages/man2/socket.2.html),
the `domain` argument to `socket` takes one of a set of constants starting with `AF_`,
such as `AF_INET`.
These constants are defined by `sys/socket.h`.
Likewise, `type` takes a constant starting with `SOCK_`, also defined by `sys/socket.h`.
The third argument, `protocol`, is typically zero.

As described in the `RETURN VALUE` section,
`socket` returns an integer, which is a file descriptor;
on error, it returns a negative number (an invalid file descriptor) and sets `errno`.

An example call in C would then be:
```c
#include <sys/socket.h>

int sock = socket(AF_INET, SOCK_STREAM, 0);
```

In rsyscall, this syscall is defined as follows:
```python
from rsyscall.sys.socket import AF, SOCK
from rsyscall import FileDescriptor

class Task:
    async def socket(self, domain: AF, type: SOCK, protocol: int=0) -> FileDescriptor:
        ...
```

The types of the arguments to `rsyscall.Task.socket`, as described in the manpage,
are represented statically in the Python type system.
The `protocol` argument defaults to zero, for convenience.

`rsyscall.sys.socket.AF` and `rsyscall.sys.socket.SOCK` are `enum.IntEnum`s;
constant values such as `AF_INET` are accessed as `rsyscall.sys.socket.AF.INET`.

These constants are imported from the "header module" `rsyscall.sys.socket`,
with the name of the module patterned after the filename of the corresponding Linux C header `sys/socket.h`.
The header module for a system call, which contains all related values,
is always the last header in the `SYNOPSIS` section of that system call's manpage.

The return type of `rsyscall.Task.socket` is also represented statically in the Python type system.
`rsyscall.Task.socket` returns a `rsyscall.FileDescriptor`;
there is no corresponding type in C,
so it's imported from the top-level `rsyscall` module.

rsyscall never returns "invalid" values;
instead, on error, `rsyscall.Task.socket` throws `OSError`,
as is typical for Python.
There is no `errno` in rsyscall, only `OSError`s thrown from syscalls.

An example call in Python with rsyscall would then be:
```python
sock = await task.socket(AF.INET, SOCK.STREAM, 0)
```

To summarize, we've followed the following rules for this translation:
- If the last header included in the `SYNOPSIS` section of the manpage for a system call is `foo/bar.h`,
  values related to that system call are present in the header module `rsyscall.foo.bar`.
- Constants are represented as `enum.IntEnums` and can be imported from the syscall's header module.
- Types (such as `rsyscall.FileDescriptor`) which have no equivalent in C are imported from `rsyscall`,
  following Python class naming rules.
- System calls throw errors instead of setting errno.

## Structs and memory
Our next example is the `bind` system call.
The [bind(2)](https://www.man7.org/linux/man-pages/man2/bind.2.html) `SYNOPSIS` section contains:

```c
#include <sys/types.h>          /* See NOTES */
#include <sys/socket.h>

int bind(int sockfd, const struct sockaddr *addr, socklen_t addrlen);
```

As described in [bind(2)](https://www.man7.org/linux/man-pages/man2/bind.2.html):
- `sockfd` is a socket file descriptor produced by the `socket` system call,
- `addr` is an initialized pointer to an address struct such as `struct sockaddr`,
- `addrlen` is the length of `addr`.

As described in the `RETURN VALUE` section,
`bind` returns zero on success;
on error, it returns a negative number and sets `errno`.

`bind(2)` mentions that to know the exact format of `addr`,
we need to refer to look at another manpage,
depending on the `AF` argument passed to the earlier `socket` call.
We'll use the socket file descriptor we made earlier with `AF.INET`,
so we'll look at the `ip(7)` manpage.

The `ip(7)` `SYNOPSIS` section contains:
```c
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h> /* superset of previous */
```

The manpage describes the definition of the `struct sockaddr_in` struct;
to see that definition, we can include `netinet/ip.h`.

The manpage mentions that the fields of `struct sockaddr_in` are in network byte order;
therefore, in C, we need to call `htons` to convert any value we store in the `.sin_port` field.

So an example call in C would be:
```c
#include <sys/socket.h>
#include <netinet/ip.h>

struct sockaddr_in addr = {
  .sin_family = AF_INET,
  .sin_port = htons(1234),
  .sin_addr = 0,
};
bind(sock, (struct sockaddr*) &addr, sizeof(addr));
```

In rsyscall, this syscall is defined as follows:
```python
from rsyscall.sys.socket import Sockaddr
from rsyscall.netinet.ip import SockaddrIn
from rsyscall import WrittenPointer

class FileDescriptor:
    async def bind(self, addr: WrittenPointer[Sockaddr]) -> None:
        ...
```

Since `rsyscall.FileDescriptor.bind` takes a file descriptor as its first argument,
it's defined as a method on `rsyscall.FileDescriptor`.

`rsyscall.sys.socket.Sockaddr` and `rsyscall.netinet.ip.SockaddrIn` are Python classes,
following Python class naming rules.
They implement `rsyscall.struct.Serializer.to_bytes` and `rsyscall.struct.Serializer.from_bytes`,
so they can be written to and read from memory as the corresponding C structs
`struct sockaddr` and `struct sockaddr_in`.
`rsyscall.netinet.ip.SockaddrIn` inherits from `rsyscall.sys.socket.Sockaddr`,
so it can be used anywhere `rsyscall.sys.socket.Sockaddr` can.

`rsyscall.WrittenPointer[rsyscall.sys.socket.Sockaddr]`
is an initialized pointer to a `rsyscall.sys.socket.Sockaddr` (or a derived type of `rsyscall.sys.socket.Sockaddr`).
Initialized pointers can be created by calling `rsyscall.Thread.ptr`,
passing a value of the appropriate type.

All `rsyscall.Pointer`s in rsyscall know their length,
so we never pass a separate length argument such as `addrlen`.

All `rsyscall.Pointer`s are garbage collected, so there's no need for manual freeing.

Since `rsyscall.FileDescriptor.bind` returns a fixed value of 0 on success,
it returns `None` in rsyscall, and like all other syscalls,
raises `OSError` to indicate an error.

An example call in Python with rsyscall would then be:
```python
addr = await thread.ptr(SockaddrIn(port=1234, addr=0))
await sock.bind(addr)
```

When creating the `rsyscall.netinet.ip.SockaddrIn`, we don't have to, nor should we, call `htons` on the port.
All rsyscall structs handle converting to and from network byte order
as part of `rsyscall.struct.Serializer.to_bytes` and `rsyscall.struct.Serializer.from_bytes`, where necessary.

The prefixes on the struct fields (the `sin_` on `sin_family`, `sin_port`, `sin_addr`) are removed;
those are [only necessary](https://stackoverflow.com/questions/10325870/why-are-the-fields-in-struct-stat-named-st-something/10325945) in C.
Positional arguments to the constructor, instead of keyword arguments, are also valid;
the argument positions are the same as the order of the fields in the struct.

The value for `family` must always be `AF_INET` for `rsyscall.netinet.ip.SockaddrIn`,
so it's set by default as a convenience for the user.

To summarize, we've added the following additional rules for this translation:
- System calls which operate on objects such as file descriptors
  are defined as methods on those objects.
- Structs passed to system calls are represented as Python classes following Python class naming rules,
  and can be imported from the appropriate header module.
- Complexities of reading and writing structures from memory, such as for network byte order, are abstracted away.
- Struct field prefixes are removed.
- Syscalls which take initialized pointers to memory take rsyscall `rsyscall.WrittenPointer`s of the appropriate type.
- `rsyscall.WrittenPointer`s can be created by passing a value of the appropriate type,
  such as Python class representations of structs, to `rsyscall.Thread.ptr`.
- Pointer length arguments are omitted in rsyscall.
* Malloc and syscalls which write to memory
Our next example is the `pipe` system call.
The [pipe(2)](https://www.man7.org/linux/man-pages/man2/pipe.2.html) `SYNOPSIS` section contains:

```c
#include <unistd.h>

int pipe(int pipefd[2]);
```

As described in [pipe(2)](https://www.man7.org/linux/man-pages/man2/pipe.2.html):
- Two file descriptors will be written to `pipefd`;
  `pipefd[0]` will contain the read end of the pipe
  and `pipefd[1]` will contain the write end of the pipe.

As described in the `RETURN VALUE` section,
`bind` returns zero on success;
on error, it returns a negative number and sets `errno`.

So an example call in C would be:
```c
#include <unistd.h>

int pipefd[2];
pipe(pipefd);
```

In rsyscall, `rsyscall.Task.pipe` is defined as follows:
```python
from rsyscall.unistd import Pipe

class Task:
    async def pipe(self, pipefd: Pointer[Pipe]) -> ReadablePointer[Pipe]:
        pass
```

Rather than an ad-hoc array, we pass a pointer to a type specific to `rsyscall.Task.pipe`, `rsyscall.unistd.Pipe`,
which will contain the two file descriptors once the `rsyscall.Task.pipe` call is done.

`rsyscall.Task.pipe` will write data to the passed-in `pipefd`,
wiping out whatever was there before,
so `pipefd` doesn't need to be initialized with data before it's passed in.
Therefore, it's a plain `rsyscall.Pointer`, not an `rsyscall.WrittenPointer`,
and we can allocate it with `rsyscall.Thread.malloc`.

The `rsyscall.Task.pipe` system call writes to the `pipefd` buffer,
so the passed-in `rsyscall.Pointer[rsyscall.unistd.Pipe]` is consumed and not usable after the call.
`rsyscall.Task.pipe` returns a new `rsyscall.ReadablePointer[rsyscall.unistd.Pipe]` for the same buffer,
from which we can read the `rsyscall.Task.pipe`.

An example call in Python with rsyscall would then be:
```python
pipefd = await thr.malloc(Pipe)
pipefd = await thr.task.pipe(pipefd)
read, write = await pipefd.read()
# or...
pipe = await pipefd.read()
assert pipe.read == pipe[0]
assert pipe.write == pipe[1]
```

To summarize, we've added the following additional rules for this translation:

- The few system calls which take arrays have types defined specifically for them, named after the system call.
- System calls which write to memory take `rsyscall.Pointer`s of the appropriate type.
- We can allocate an uninitialized `rsyscall.Pointer`
  by passing a type and (when appropriate) a size to `rsyscall.Thread.malloc`.
- System calls which write to memory consume the `rsyscall.Pointer`s that are passed in,
  and return one or more new `rsyscall.ReadablePointer`s for the readable portion of the passed-in buffers.

## Bitflags and syscalls which return a size
Our next and final example is the `recv` system call.
The [recv(2)](https://www.man7.org/linux/man-pages/man2/recv.2.html) `SYNOPSIS` section contains:
```c
#include <sys/types.h>
#include <sys/socket.h>

ssize_t recv(int sockfd, void *buf, size_t len, int flags);
```

As described in [recv(2)](https://www.man7.org/linux/man-pages/man2/recv.2.html):
- `sockfd` is a socket file descriptor
- `buf` is a pointer to some memory
- `len` is the maximum number of bytes which will be received from `sockfd` and written to `buf`
- `flags` is a bitflag, created by or-ing together constants starting with `MSG_`

As described in the `RETURN VALUE` section,
`recv` returns the number of bytes received and written to `buf`, which may be less than `len`.
On error, it returns a negative number and sets `errno`.

An example call in C would be:
```c
#include <sys/socket.h>

char buf[4096];
ssize_t size = recv(fd, buf, sizeof(buf), MSG_DONTWAIT|MSG_PEEK);
```

The user would then examine the first `size` bytes of `buf`;
the last `sizeof(buf) - size` bytes are uninitialized and invalid to load.

In rsyscall, this syscall is defined as follows:
```python
from rsyscall.sys.socket import MSG
from rsyscall import Pointer, ReadablePointer

class FileDescriptor:
    async def recv(self, buf: Pointer[bytes], flags: MSG) -> (ReadablePointer[bytes], Pointer[bytes]):
        ...
```

Like `rsyscall.FileDescriptor.bind`, since `rsyscall.FileDescriptor.recv` takes a file descriptor as its first argument,
it's defined as a method on `rsyscall.FileDescriptor`.

We pass a `rsyscall.Pointer[bytes]` as our buffer.
We can allocate one of these with `rsyscall.Thread.malloc`.
We pass an appropriate type and size to allocate to `rsyscall.Thread.malloc`,
and it returns a `rsyscall.Pointer` of that type and with that size.

As mentioned previously, `rsyscall.Pointer`s know their length,
so we don't need to pass `count` to `rsyscall.FileDescriptor.recv`.
If we want to pass a smaller value for `count`,
we can make a smaller buffer.

The `flags` argument, as mentioned previously, is typed as a Python `enum` class;
in this case, since it is a bitflag, it is an `enum.IntFlag`, which is combinable with bitwise operators.

To preserve type-safety of `rsyscall.FileDescriptor.recv` operations,
`rsyscall.FileDescriptor.recv` does not return a size.
Instead, it consumes `buf` and splits it into two new buffers, which it returns in a tuple.
This applies to all system calls which return a size,
such as `rsyscall.FileDescriptor.send`, `rsyscall.FileDescriptor.read`, `rsyscall.FileDescriptor.write`, and others.

The first element of the tuple is readable;
it points to the data which was received from the socket,
and which can now be read from memory with `rsyscall.ReadablePointer.read`.

The second element of the tuple is not readable;
it is the leftover part of the original buffer,
which was not filled with data from the socket.

We can check how much data was read by checking the size of the first pointer with `rsyscall.Pointer.size`.

An example call in Python with rsyscall would then be:
```python
from rsyscall.sys.socket import MSG

buf = await thread.malloc(bytes, 4096)
received, leftover = await sock.recv(buf, MSG.DONTWAIT|MSG.PEEK)
data = await received.read()
```

After the `rsyscall.FileDescriptor.recv` call, `buf` can no longer be used.
We can merge `received` and `leftover` back together with `rsyscall.Pointer.merge`:
`received.merge(leftover)`, or `received + leftover` for short.
Then we can reuse the resulting buffer.

To summarize, we've added the following additional rules:
- System calls returning sizes which offset into passed-in buffers return tuples of pointers:
  The first pointer points to the range from the start of the buffer to the size;
  the second pointer points to the range from the size to the end of the buffer.
  These two pointers will be different types depending on what operations are valid on each part of the buffer.

## Additional notes
### Everything is garbage collected
`rsyscall.FileDescriptor`s, `rsyscall.sys.mman.MemoryMapping`s, `rsyscall.handle.ChildProcess`es, and other resources
are all allocated through various system calls which return Python objects.
System calls related to these resources are present as methods on these objects.

rsyscall handles closing these resources once the last reference goes out of scope.
For example, when the last reference to a file descriptor goes out of scope, the file descriptor is closed.
The same applies for memory mappings, child processes, and other objects.

It is also possible to manually close an `rsyscall.FileDescriptor`
(or unmap an `rsyscall.sys.mman.MemoryMapping`, or kill a `rsyscall.handle.ChildProcess`)
if you want it to be deterministically closed.

### syscalls valid on multiple objects
Some syscalls can operate on multiple kinds of objects;
for example, `setpriority` can both operate on the current process and other processes.
In such cases, all the objects which are valid targets for the syscall will have a method for the syscall;
there's both a `rsyscall.Task.setpriority` and a `rsyscall.handle.Process.setpriority`.

### Deviations from Linux headers
In rare cases, to improve type safety and usability,
we will intentionally deviate from how the Linux headers define things.

Unfortunately, for such APIs you must resort to the rsyscall documentation to understand their usage.
We therefore avoid this wherever possible, but it's sometimes necessary.

`struct msghdr` is one example;
it is used differently by `rsyscall.FileDescriptor.sendmsg` and `rsyscall.FileDescriptor.recvmsg`,
and in the case of `rsyscall.FileDescriptor.recvmsg` also serves as an out-parameter.
To preserve type-safety,
it is represented with three classes,
`rsyscall.sys.socket.SendMsghdr`, `rsyscall.sys.socket.RecvMsghdr`, and `rsyscall.sys.socket.RecvMsghdrOut`.

### Missing syscalls
We seek to provide user-accessible low-level interfaces to the entirety of the Linux kernel,
including all non-obsolete syscalls,
and including things that are typically considered low-level implementation details (for example, `rsyscall.linux.futex`).

If you want to use some feature of the Linux kernel that is missing an interface in rsyscall,
we're happy to add support for it, just file an issue.
"""



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
from rsyscall.tests.utils import do_async_things
from rsyscall.epoller import Epoller
from rsyscall.monitor import AsyncSignalfd

from rsyscall.sched import CLONE
from rsyscall.signal import SIG, Sigset
from rsyscall.stdlib import mkdtemp
from rsyscall.sys.signalfd import SignalfdSiginfo
from rsyscall.sys.wait import CalledProcessError

class TestClone(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = await local_thread.clone(CLONE.FILES)

    async def test_exit(self) -> None:
        await self.thr.exit(0)

    async def test_nest_exit(self) -> None:
        thread = await self.thr.clone(CLONE.FILES)
        await thread.exit(0)

    async def test_nest_multiple(self) -> None:
        for i in range(5):
            child = await self.thr.clone()
            await do_async_things(self, child.epoller, child)
            await child.exit(0)

    async def test_two_children_exec(self) -> None:
        """Start two child and exec in each of them.

        This test would (sometimes) catch a race condition we had where waitpid
        on one child would consume the SIGCHLD for another child, and our logic
        to eagerly call waitid was wrong, so waitpid on the other child would
        block forever.

        """
        thr2 = await local_thread.clone()
        cmd = self.thr.environ.sh.args('-c', 'true')
        child1 = await self.thr.exec(cmd)
        child2 = await thr2.exec(cmd)
        await child1.check()
        await child2.check()

    async def test_async(self) -> None:
        epoller = await Epoller.make_root(self.thr.ram, self.thr.task)
        await do_async_things(self, epoller, self.thr)

    async def test_nest_async(self) -> None:
        thread = await self.thr.clone(CLONE.FILES)
        epoller = await Epoller.make_root(thread.ram, thread.task)
        await do_async_things(self, epoller, thread)
        await thread.exit(0)

    async def test_unshare_async(self) -> None:
        await self.thr.unshare(CLONE.FILES)
        thread = await self.thr.clone(CLONE.FILES)
        epoller = await Epoller.make_root(thread.ram, thread.task)
        await thread.unshare(CLONE.FILES)
        await do_async_things(self, epoller, thread)
        await thread.exit(0)

    async def test_exec(self) -> None:
        child = await self.thr.exec(self.thr.environ.sh.args('-c', 'false'))
        with self.assertRaises(CalledProcessError):
            await child.check()

    async def test_check_in_nursery(self) -> None:
        "We broke this with some concurrency refactoring once"
        child = await self.thr.exec(self.thr.environ.sh.args('-c', 'sleep inf'))
        self.nursery.start_soon(child.check)

    async def test_nest_exec(self) -> None:
        child = await self.thr.clone()
        grandchild = await child.clone()
        cmd = self.thr.environ.sh.args('-c', 'true')
        await (await child.exec(cmd)).check()
        await (await grandchild.exec(cmd)).check()

    async def test_mkdtemp(self) -> None:
        async with (await mkdtemp(self.thr)):
            pass

    async def test_signal_queue(self) -> None:
        epoller = await Epoller.make_root(self.thr.ram, self.thr.task)
        sigfd = await AsyncSignalfd.make(self.thr.ram, self.thr.task, epoller, Sigset({SIG.INT}))
        sigevent = sigfd.next_signal
        await self.thr.process.kill(SIG.INT)
        await sigevent.wait()

class TestCloneUnshareFiles(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.local = local_thread
        self.thr = await self.local.clone()

    async def asyncTearDown(self) -> None:
        await self.thr.exit(0)

    async def test_nest_async(self) -> None:
        thread = await self.thr.clone()
        epoller = await Epoller.make_root(thread.ram, thread.task)
        await do_async_things(self, epoller, thread)
        await thread.exit(0)



"`#include <sys/mman.h>`"
from __future__ import annotations
from rsyscall._raw import lib # type: ignore
from rsyscall.linux.memfd import MFD
import enum

__all__ = [
    "MFD",
    "PROT",
    "MAP",
    "MappableFileDescriptor",
    "MemoryMappingTask",
    "MemoryMapping",
]

class PROT(enum.IntFlag):
    EXEC = lib.PROT_EXEC
    READ = lib.PROT_READ
    WRITE = lib.PROT_WRITE
    NONE = lib.PROT_NONE

class MAP(enum.IntFlag):
    PRIVATE = lib.MAP_PRIVATE
    SHARED = lib.MAP_SHARED
    ANONYMOUS = lib.MAP_ANONYMOUS
    POPULATE = lib.MAP_POPULATE

#### Classes ####
from dataclasses import dataclass
import rsyscall.far
import rsyscall.near
import rsyscall.near as near
import typing as t

@dataclass(frozen=True)
class MemoryMapping:
    __slots__ = ('task', 'near', 'file')
    task: MemoryMappingTask
    near: rsyscall.near.MemoryMapping
    file: rsyscall.far.File

    async def munmap(self) -> None:
        await _munmap(self.task.sysif, self.near)

    def for_task(self, task: MemoryMappingTask) -> MemoryMapping:
        if task.address_space != self.task.address_space:
            raise rsyscall.far.AddressSpaceMismatchError()
        return MemoryMapping(task, self.near, self.file)

    def __str__(self) -> str:
        return f"MemoryMapping({str(self.task)}, {str(self.near)})"

    def __repr__(self) -> str:
        return str(self)

from rsyscall.handle.fd import BaseFileDescriptor, FileDescriptorTask

class MemoryMappingTask(FileDescriptorTask):
    async def mmap(self, length: int, prot: PROT, flags: MAP,
                   page_size: int=4096,
    ) -> MemoryMapping:
        # a mapping without a file descriptor, is an anonymous mapping
        flags |= MAP.ANONYMOUS
        ret = await _mmap(self.sysif, length, prot, flags, page_size=page_size)
        return MemoryMapping(self, ret, rsyscall.far.File())

class MappableFileDescriptor(BaseFileDescriptor):
    task: MemoryMappingTask

    async def mmap(self, length: int, prot: PROT, flags: MAP,
                   offset: int=0,
                   page_size: int=4096,
                   file: rsyscall.far.File=None,
    ) -> MemoryMapping:
        self._validate()
        if file is None:
            file = rsyscall.far.File()
        ret = await _mmap(self.task.sysif, length, prot, flags, fd=self.near, offset=offset, page_size=page_size)
        return MemoryMapping(self.task, ret, file)

#### Raw syscalls ####
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _mmap(sysif: SyscallInterface, length: int, prot: PROT, flags: MAP,
               addr: t.Optional[near.Address]=None,
               fd: t.Optional[near.FileDescriptor]=None, offset: int=0,
               page_size: int=4096) -> near.MemoryMapping:
    if addr is None:
        addr = 0 # type: ignore
    else:
        assert (int(addr) % page_size) == 0
    if fd is None:
        fd = -1 # type: ignore
    # TODO we want Linux to enforce this for us, but instead it just rounds,
    # leaving us unable to later munmap.
    assert (int(length) % page_size) == 0
    ret = await sysif.syscall(SYS.mmap, addr, length, prot, flags, fd, offset)
    return near.MemoryMapping(address=ret, length=length, page_size=page_size)

async def _munmap(sysif: SyscallInterface, mapping: near.MemoryMapping) -> None:
    await sysif.syscall(SYS.munmap, mapping.address, mapping.length)



"""Functions and classes relating to Unix environment variables.

This is always available to us, and we can use it to look up executables in our
environment and access various other resources.

This will be superseded in many situations, though; for example, if we have
access to a Nix store, we do not need to look at the environment to find
executables.

"""
from __future__ import annotations
from dneio import run_all
from rsyscall.command import Command
from rsyscall.handle import Task, FileDescriptor, WrittenPointer
from rsyscall.memory.ram import RAM
from rsyscall.path import Path
from rsyscall.unistd import ArgList
import os
import typing as t
import functools

from rsyscall.fcntl import O
from rsyscall.unistd import OK
__all__ = [
    'ExecutableNotFound',
    'Environment',
]

T = t.TypeVar('T')
def chunks(lst: t.List[T], size: int) -> t.Iterator[t.List[T]]:
    """Yield chunks of `lst`, at most `size` long.

    Thanks Stack Overflow

    """
    for i in range(0, len(lst), size):
        yield lst[i:i+size]

class ExecutableNotFound(Exception):
    "No executable with this name can be found; thrown from `Environment.which`"
    def __init__(self, name: str) -> None:
        super().__init__(name)
        self.name = name

class ExecutablePathCache:
    "A cache of executables looked up on PATH."
    def __init__(self, task: Task, ram: RAM, paths: t.List[str]) -> None:
        self.task = task
        self.ram = ram
        self.paths = [Path(path) for path in paths]
        self.fds: t.Dict[Path, t.Optional[FileDescriptor]] = {}
        self.path_cache: t.Dict[str, Path] = {}

    async def _get_fd_for_path(self, path: Path) -> t.Optional[FileDescriptor]:
        "Return a cached file descriptor for this path."
        if path not in self.fds:
            try:
                fd = await self.task.open(await self.ram.ptr(path), O.PATH|O.DIRECTORY)
            except OSError:
                self.fds[path] = None
            else:
                self.fds[path] = fd
        return self.fds[path]

    async def _check(self, path: Path, name: WrittenPointer[str]) -> bool:
        "Return true if there's an executable with this name under this path"
        fd = await self._get_fd_for_path(path)
        if fd is None:
            return False
        else:
            try:
                # TODO hmm this returns fine for directories tho. hm. hm.
                # oh well we'll just fail at exec time, that was always possible
                await fd.faccessat(name, OK.X)
            except OSError:
                return False
            else:
                return True

    async def which(self, name: str) -> Command:
        "Locate an executable with this name on PATH; throw ExecutableNotFound on failure"
        try:
            path = self.path_cache[name]
        except KeyError:
            nameptr = await self.ram.ptr(name)
            # do the lookup for 64 paths at a time, that seems like a good batching number
            for paths in chunks(self.paths, 64):
                thunks = [functools.partial(self._check, path, nameptr) for path in paths]
                results = await run_all(thunks) # type: ignore
                for path, result in zip(paths, results):
                    if result:
                        # path is set as the loop variable; python has no scope
                        # so we can just break out and use it.
                        break
                if result:
                    break
            else:
                raise ExecutableNotFound(name)
            self.path_cache[name] = path
        return Command(path/name, [name], {})

__pdoc__ = {
    'Environment.__getitem__': True,
    'Environment.__contains__': True,
    'Environment.__len__': True,
    'Environment.__delitem__': True,
    'Environment.__setitem__': True,
}
class Environment:
    "A representation of Unix environment variables."
    @staticmethod
    def make_from_environ(task: Task, ram: RAM, environment: t.Dict[str, str]) -> Environment:
        return Environment(environment, ExecutablePathCache(task, ram, environment.get("PATH", "").split(":")))

    def __init__(self, environment: t.Dict[str, str], path: ExecutablePathCache,
                 arglist_ptr: t.Optional[WrittenPointer[ArgList]]=None,
    ) -> None:
        self.data = environment
        self.sh = Command(Path("/bin/sh"), ['sh'], {})
        "The POSIX-required `/bin/sh`, as a `rsyscall.Command`"
        self.tmpdir = Path(self.get("TMPDIR", "/tmp"))
        "`TMPDIR`, or `/tmp` if it's not set, as a `rsyscall.path.Path`"
        self.path = path
        self.arglist_ptr = arglist_ptr

    def __getitem__(self, key: str) -> str:
        return self.data[key]

    def __contains__(self, key: str) -> bool:
        return key in self.data

    def __len__(self) -> int:
        return len(self.data)

    def __delitem__(self, key: str) -> None:
        del self.data[key]

    def __setitem__(self, key: str, val: str) -> None:
        self.data[key] = val

    def get(self, key: str, default: str) -> str:
        "Like `dict.get`; get an environment variable, with a default."
        result = self.data.get(key)
        if result is None:
            return default
        else:
            return result

    async def which(self, name: str) -> Command:
        "Locate an executable with this name on `PATH`; throw `ExecutableNotFound` on failure."
        return await self.path.which(name)

    def inherit(self, task: Task, ram: RAM) -> Environment:
        """Return a new Environment instance for this Task and RAM.

        We share the existing ExecutablePathCache. This centralizes path lookups so that
        they're shared between all threads.

        """
        return Environment(dict(self.data), self.path, self.arglist_ptr)

    async def as_arglist(self, ram: RAM) -> WrittenPointer[ArgList]:
        if self.arglist_ptr is None:
            envp = ['='.join([key, value]) for key, value in self.data.items()]
            async def op(sem: RAM) -> WrittenPointer[ArgList]:
                envp_ptrs = ArgList([await sem.ptr(arg) for arg in envp])
                return await sem.ptr(envp_ptrs)
            ptr: WrittenPointer[ArgList] = await ram.perform_batch(op)
            self.arglist_ptr = ptr
        return self.arglist_ptr



from __future__ import annotations
from rsyscall._raw import ffi, lib # type: ignore
import os
import typing as t
import enum

class MFD(enum.IntFlag):
    NONE = 0
    CLOEXEC = lib.MFD_CLOEXEC
    ALLOW_SEALING = lib.MFD_ALLOW_SEALING
    HUGETLB = lib.MFD_HUGETLB
    HUGE_2MB = lib.MFD_HUGE_2MB
    HUGE_1GB = lib.MFD_HUGE_1GB
    
#### Classes ####
from rsyscall.handle.fd import T_fd, FileDescriptorTask
if t.TYPE_CHECKING:
    from rsyscall.handle.pointer import WrittenPointer

class MemfdTask(FileDescriptorTask[T_fd]):
    async def memfd_create(self, name: WrittenPointer[t.Union[str, os.PathLike]], flags: MFD=MFD.NONE) -> T_fd:
        with name.borrow(self) as name_n:
            fd = await _memfd_create(self.sysif, name_n, flags|MFD.CLOEXEC)
            return self.make_fd_handle(fd)

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _memfd_create(sysif: SyscallInterface,
                        name: near.Address, flags: MFD) -> near.FileDescriptor:
    return near.FileDescriptor(await sysif.syscall(SYS.memfd_create, name, flags))



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
from rsyscall.sys.socket import AF, SOCK, Socketpair
from rsyscall.unistd import Pipe
from rsyscall.fcntl import O

from rsyscall.sched import CLONE
from rsyscall.tests.utils import assert_thread_works

class TestPidns(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.local = local_thread
        self.init = await self.local.clone(CLONE.NEWUSER|CLONE.NEWPID|CLONE.FILES)

    async def test_cat(self) -> None:
        cat = await self.local.environ.which('cat')
        pair = await (await self.local.task.socketpair(
            AF.UNIX, SOCK.STREAM, 0, await self.local.ram.malloc(Socketpair))).read()
        child = await self.init.clone()
        child_side = child.task.inherit_fd(pair.first)
        # close in parent so we'll get EOF on other side when cat dies
        await pair.first.close()
        await child_side.dup2(child.stdin)
        await child_side.dup2(child.stdout)
        child_process = await child.exec(cat)
        await self.init.exit(0)
        # cat dies, get EOF on socket
        read, _ = await pair.second.read(await self.local.ram.malloc(bytes, 16))
        self.assertEqual(read.size(), 0)

    async def test_sleep(self) -> None:
        pipe = await (await self.local.task.pipe(await self.local.ram.malloc(Pipe))).read()
        child = await self.init.clone()
        child_fd = child.task.inherit_fd(pipe.write)
        await pipe.write.close()
        await child_fd.disable_cloexec()
        child_process = await child.exec(child.environ.sh.args('-c', '{ sleep inf & } &'))
        await child_process.check()
        await self.init.exit(0)
        read, _ = await pipe.read.read(await self.local.ram.malloc(bytes, 1))
        self.assertEqual(read.size(), 0)



"""Several REPL-based `wish.WishGranter`s for the `wish` library

This module provides several choices of `wish.WishGranter`s for the wish library. Each
`WishGranter` is an asynchronous REPL based on the `arepl` library.

In fact, this library automatically instantiates `ConsoleGenie` and sets
`wish.my_wish_granter` to it at import time. Not very nice, but convenient.

Also, for convenience, we re-export `wish.wish` and `wish.Wish`.

"""
from __future__ import annotations
from rsyscall.command import Command
from rsyscall.epoller import AsyncFileDescriptor
from rsyscall.thread import Thread
from rsyscall.path import Path
from wish import wish, WishGranter, Wish, my_wish_granter
import logging
import os
import arepl
import sys
import traceback
import trio
import typing as t

from rsyscall.sched import CLONE
from rsyscall.sys.socket import SOCK, AF
from rsyscall.sys.un import SockaddrUn
from rsyscall.sys.wait import W
from rsyscall.unistd import Pipe

logger = logging.getLogger(__name__)

__all__ = [
    'ConsoleGenie',
    'ConsoleServerGenie',
    'run_repl',
    'serve_repls',
]

T = t.TypeVar('T')

class ConsoleGenie(WishGranter):
    "A WishGranter which satisfies wishes by starting a REPL on stdin/stdout for human intervention"
    @classmethod
    async def make(self, thread: Thread):
        "Create a ConsoleGenie that will serve using `thread`'s stdin/stdout"
        cat = await thread.environ.which("cat")
        return ConsoleGenie(thread, cat)

    def __init__(self, thread: Thread, cat: Command) -> None:
        self.thread = thread
        self.cat = cat
        self.lock = trio.Lock()

    async def wish(self, wish: Wish[T]) -> T:
        """Serve a REPL on stdin/stdout and returns the value returned from it; throws on REPL hangup

        We use cat as a hack to be async; we can't robustly turn stdin/stdout into
        AsyncFDs, because setting them to NONBLOCK will be inherited by other processes
        using the same terminal. Instead, we just start two cats to do blocking
        reads/writes of stdin/stdout, then do async reads/writes to the cats.

        """
        async with self.lock:
            message = "".join(traceback.format_exception(None, wish, wish.__traceback__))
            wisher_frame = [frame for (frame, lineno) in traceback.walk_tb(wish.__traceback__)][-1]

            to_term_pipe = await (await self.thread.task.pipe(await self.thread.ram.malloc(Pipe))).read()
            from_term_pipe = await (await self.thread.task.pipe(await self.thread.ram.malloc(Pipe))).read()
            async_from_term = await self.thread.make_afd(from_term_pipe.read, set_nonblock=True)
            async_to_term = await self.thread.make_afd(to_term_pipe.write, set_nonblock=True)
            try:
                cat_stdin_thread = await self.thread.clone()
                await cat_stdin_thread.task.inherit_fd(to_term_pipe.read).dup2(cat_stdin_thread.stdin)
                async with await cat_stdin_thread.exec(self.cat):
                    cat_stdout_thread = await self.thread.clone()
                    await cat_stdout_thread.task.inherit_fd(from_term_pipe.write).dup2(cat_stdout_thread.stdout)
                    async with await cat_stdout_thread.exec(self.cat):
                        ret = await run_repl(async_from_term, async_to_term, {
                            **wisher_frame.f_locals,
                            **wisher_frame.f_globals,
                            '__repl_stdin__': async_from_term,
                            '__repl_stdout__': async_to_term,
                            'wish': wish,
                            'wisher_frame': wisher_frame,
                        }, wish.return_type, message)
                        return ret
            finally:
                await async_from_term.close()
                await async_to_term.close()

class ConsoleServerGenie(WishGranter):
    """On wish, listens for connections to a socket in `sockdir`, and serves a REPL to them

    Also, immediately starts socat from `thread` connecting to the socket; that means a
    REPL is served on stdin/stdout.

    The socket names are generated from the function and line number at which wish was
    called. The sockets are unlinked after a value is returned from some REPL.

    Multiple connections to the socket are supported at a time. Once one returns a value,
    all are cancelled.

    """
    @classmethod
    async def make(self, thread: Thread, sockdir: Path):
        socat = await thread.environ.which("socat")
        return ConsoleServerGenie(thread, sockdir, socat)

    def __init__(self, thread: Thread, sockdir: Path, socat: Command) -> None:
        self.thread = thread
        self.sockdir = sockdir
        self.socat = socat
        self.name_counts: t.Dict[str, int] = {}

    def _uniquify_name(self, name: str) -> str:
        "We never reuse a name for a socket."
        if name not in self.name_counts:
            self.name_counts[name] = 1
            return name
        else:
            self.name_counts[name] += 1
            return name + f".{self.name_counts[name]}"

    async def wish(self, wish: Wish[T]) -> T:
        message = "".join(traceback.format_exception(None, wish, wish.__traceback__))
        wisher_frame = [frame for (frame, lineno) in traceback.walk_tb(wish.__traceback__)][-1]
        sock_name = self._uniquify_name(f'{wisher_frame.f_code.co_name}-{wisher_frame.f_lineno}')
        sock_path = self.sockdir/sock_name
        cmd = self.socat.args("-", "UNIX-CONNECT:" + os.fsdecode(sock_path))
        sockfd = await self.thread.make_afd(await self.thread.socket(AF.UNIX, SOCK.STREAM|SOCK.NONBLOCK))
        await sockfd.bind(await self.thread.ptr(await SockaddrUn.from_path(self.thread, sock_path)))
        await sockfd.handle.listen(10)
        async with trio.open_nursery() as nursery:
            @nursery.start_soon
            async def do_socat():
                while True:
                    thread = await self.thread.clone()
                    try:
                        child = await thread.exec(cmd)
                    except:
                        await thread.exit(0)
                        raise
                    async with child:
                        await child.waitpid(W.EXITED)
            ret = await serve_repls(sockfd, {
                **wisher_frame.f_locals,
                **wisher_frame.f_globals,
                'wisher_frame': wisher_frame,
            }, wish.return_type, message)
            nursery.cancel_scope.cancel()
        await self.thread.task.unlink(await self.thread.ram.ptr(sock_path))
        return ret

async def run_repl(infd: AsyncFileDescriptor,
                   outfd: AsyncFileDescriptor,
                   global_vars: t.Dict[str, t.Any],
                   wanted_type: t.Type[T], message: str) -> T:
    """Serve a REPL on infd/outfd, guarding that it only returns `wanted_type`

    We print `message` on startup.  `global_vars` is mutably used as the global variables
    dict for the REPL.

    """
    async with trio.open_nursery() as repl_nursery:
        @repl_nursery.start_soon
        async def wait_for_rdhup() -> None:
            await infd.wait_for_rdhup()
            # when we get RDHUP on the connection, we want to cancel the REPL, even if
            # some task is in progress.
            raise Exception("REPL connection hangup")
        await outfd.write_all_bytes((message+"\n").encode())
        ret = await arepl.run_repl(infd.read_some_bytes, outfd.write_all_bytes, global_vars, wanted_type)
        repl_nursery.cancel_scope.cancel()
    return ret

async def serve_repls(listenfd: AsyncFileDescriptor,
                      initial_vars: t.Dict[str, t.Any],
                      wanted_type: t.Type[T], message: str) -> T:
    """Serve REPLs on a socket until someone gives us the type we want

    Multiple connections to the socket can be active simultaneously, with REPLs being
    served to all.

    Each REPL has a reference to the REPL variables of all other REPLs, through the
    __repls__ variable present in each REPL.

    A good extension would be to support killing a REPL from another REPL. We didn't do
    that for some reason, not sure why.

    """
    repl_vars: t.Dict[str, t.Dict[str, t.Any]] = {}
    retval = None
    async with trio.open_nursery() as nursery:
        async def do_repl(connfd: AsyncFileDescriptor,
                          global_vars: t.Dict[str, t.Any]) -> None:
            try:
                ret = await run_repl(connfd, connfd, global_vars, wanted_type, message)
            except arepl.FromREPL as e:
                raise e.exn from e
            except Exception:
                logger.exception("run_repl's internal logic raised an exception, disconnecting that REPL and continuing")
            else:
                nonlocal retval
                retval = ret
                nursery.cancel_scope.cancel()
            finally:
                await connfd.close()
        num = 0
        while True:
            connfd = await listenfd.make_new_afd(await listenfd.accept(SOCK.NONBLOCK))
            global_vars = {**initial_vars,
                           '__repls__': repl_vars,
                           '__repl_stdin__': connfd,
                           '__repl_stdout__': connfd}
            repl_vars[str(num)] = global_vars
            nursery.start_soon(do_repl, connfd, global_vars)
            num += 1
    return retval

def _initialize_module() -> None:
    from rsyscall import local_thread
    my_wish_granter.set(trio.run(ConsoleGenie.make, local_thread))
_initialize_module()



"""Functions and classes for a connection between two threads, with which we can open channels for data transfer
"""
from __future__ import annotations
from dneio import make_n_in_parallel
import abc
import typing as t
import trio
from rsyscall.epoller import AsyncFileDescriptor, Epoller
from rsyscall.handle import FileDescriptor, WrittenPointer, Task
from rsyscall.memory.ram import RAM

from rsyscall.sys.socket import AF, SOCK, Sockaddr, SendmsgFlags, RecvmsgFlags, SendMsghdr, RecvMsghdr, CmsgList, CmsgSCMRights, Socketpair
from rsyscall.sys.uio import IovecList
from rsyscall.fcntl import F, O

class Connection:
    """A connection between two threads through which more bidirectional channels can be opened

    You could think of this as a pre-established route between two threads; or a cable
    between them; these are connections which can be multiplexed over to create multiple
    bidirectional channels for data transfer.

    This is not necessarily a connection in the style of TCP as such; it merely represents
    that there is a way to open channels between two threads, not that there is any active
    transfer between them at the moment. TCP doesn't support opening new channels, so
    merely having an open TCP connection isn't enough to implement this interface. On the
    other hand, SCTP and QUIC do support opening new channels, so a SCTP or QUIC
    connection would be enough to implement this interface.

    """
    @abc.abstractmethod
    async def open_channels(self, count: int) -> t.List[t.Tuple[FileDescriptor, FileDescriptor]]:
        "Batched version of open_channel"
        pass

    async def open_channel(self) -> t.Tuple[FileDescriptor, FileDescriptor]:
        """Open a bidirectional channel between the two threads inside this connection

        The left side of a channel is the "local" side, and the right side is the "remote"
        side. Accesses to the local side are typically more efficient than accesses to the
        right side. Typically, in fact, the left side of the channel is in the local
        thread, although this is not required to be true.

        """
        [pair] = await self.open_channels(1)
        return pair

    @abc.abstractmethod
    async def open_async_channels(self, count: int) -> t.List[t.Tuple[AsyncFileDescriptor, FileDescriptor]]:
        "Batched version of open_async_channel"
        pass

    async def open_async_channel(self) -> t.Tuple[AsyncFileDescriptor, FileDescriptor]:
        """Like open_channel, but returns the left side as an AsyncFileDescriptor

        As discussed in open_channel's docstring, the left side is the "local" side, and
        is more efficient to access. Here we take that further: Since an AFD comes with a
        RAM, we can immediately read and write Python bytes to the left side of this
        channel. This provides an efficient way to transfer data between Python and things
        which operate on file descriptors, such as subprocesses.

        This is how we establish all our syscall or data connections for new threads;
        open_async_channel returns the most direct path possible, so we know that when
        we're sending syscalls to a thread, that data is being efficiently transferred; in
        most cases, directly between the local thread and the syscall server.

        """
        [pair] = await self.open_async_channels(1)
        return pair


    @abc.abstractmethod
    async def prep_fd_transfer(self) -> t.Tuple[FileDescriptor, t.Callable[[Task, RAM, FileDescriptor], Connection]]:
        """Prepare to transfer this Connection to another task; call the callable to execute

        The user should use whatever means to transport the returned file descriptor to
        the other task, then call the callable with the appropriate other task, a RAM for
        it, and the transferred file descriptor.

        The user might do this through fd inheritance, or maybe passing the fd over a Unix
        socket with SCM_RIGHTS.

        This is an async method because the connection might need to allocate some
        resources to do this.

        """
        pass

    @abc.abstractmethod
    def inherit(self, task: Task, ram: RAM) -> Connection:
        """Transfer this Connection to a new task by using "inherit" to move the fds"""
        pass

class FDPassConnection(Connection):
    """A socketpair between two threads over which we can pass fds to establish new channels

    If the two threads are in the same fd table, then we'll skip using SCM_RIGHTS to pass
    the new socketpairs around and instead just change the fd owner.

    See Connnection for more details on this interface.

    """
    @staticmethod
    async def make(task: Task, ram: RAM, epoller: Epoller) -> FDPassConnection:
        pair = await (await task.socketpair(AF.UNIX, SOCK.STREAM, 0, await ram.malloc(Socketpair))).read()
        return FDPassConnection(task, ram, epoller, pair.first, task, ram, pair.second)

    def __init__(self, access_task: Task, access_ram: RAM, access_epoller: Epoller, access_fd: FileDescriptor,
                 task: Task, ram: RAM, fd: FileDescriptor) -> None:
        self.access_task = access_task
        self.access_ram = access_ram
        self.access_epoller = access_epoller
        self.access_fd = access_fd
        self.task = task
        self.ram = ram
        self.fd = fd

    async def move_fds(self, fds: t.List[FileDescriptor]) -> t.List[FileDescriptor]:
        "Move the passed-in file descriptors from self.access_task to self.task"
        if self.access_task.fd_table == self.task.fd_table:
            return [fd.move(self.task) for fd in fds]
        async def sendmsg_op(sem: RAM) -> WrittenPointer[SendMsghdr]:
            iovec = await sem.ptr(IovecList([await sem.malloc(bytes, 1)]))
            cmsgs = await sem.ptr(CmsgList([CmsgSCMRights([fd for fd in fds])]))
            return await sem.ptr(SendMsghdr(None, iovec, cmsgs))
        _, [] = await self.access_fd.sendmsg(await self.access_ram.perform_batch(sendmsg_op))
        async def recvmsg_op(sem: RAM) -> WrittenPointer[RecvMsghdr]:
            iovec = await sem.ptr(IovecList([await sem.malloc(bytes, 1)]))
            cmsgs = await sem.ptr(CmsgList([CmsgSCMRights([fd for fd in fds])]))
            return await sem.ptr(RecvMsghdr(None, iovec, cmsgs))
        _, [], hdr = await self.fd.recvmsg(await self.ram.perform_batch(recvmsg_op))
        cmsgs_ptr = (await hdr.read()).control
        if cmsgs_ptr is None:
            raise Exception("cmsgs field of header is, impossibly, None")
        [cmsg] = await cmsgs_ptr.read()
        if not isinstance(cmsg, CmsgSCMRights):
            raise Exception("expected SCM_RIGHTS cmsg, instead got", cmsg)
        passed_socks = cmsg
        for sock in fds:
            await sock.close()
        return passed_socks

    async def open_channels(self, count: int) -> t.List[t.Tuple[FileDescriptor, FileDescriptor]]:
        async def make() -> Socketpair:
            return await (await self.access_task.socketpair(
                AF.UNIX, SOCK.STREAM, 0, await self.access_ram.malloc(Socketpair))).read()
        pairs = await make_n_in_parallel(make, count)
        fds = await self.move_fds([pair.second for pair in pairs])
        return [(pair.first, fd) for pair, fd in zip(pairs, fds)]

    async def open_async_channels(self, count: int) -> t.List[t.Tuple[AsyncFileDescriptor, FileDescriptor]]:
        chans = await self.open_channels(count)
        access_socks, local_socks = zip(*chans)
        async def make_afd(sock: FileDescriptor) -> AsyncFileDescriptor:
            # have to set NONBLOCK after creation because we want the other end to be blocking
            await sock.fcntl(F.SETFL, O.NONBLOCK)
            return await AsyncFileDescriptor.make(self.access_epoller, self.access_ram, sock)
        async_access_socks = [await make_afd(sock) for sock in access_socks]
        return list(zip(async_access_socks, local_socks))

    async def prep_fd_transfer(self) -> t.Tuple[FileDescriptor, t.Callable[[Task, RAM, FileDescriptor], FDPassConnection]]:
        return self.fd, self.for_task_with_fd

    def for_task_with_fd(self, task: Task, ram: RAM, fd: FileDescriptor) -> FDPassConnection:
        return FDPassConnection(
            self.access_task,
            self.access_ram,
            self.access_epoller,
            self.access_fd,
            task, ram, fd)

    def inherit(self, task: Task, ram: RAM) -> FDPassConnection:
        return self.for_task_with_fd(task, ram, self.fd.inherit(task))

class ListeningConnection(Connection):
    """An (address, listening socket) pair with which we can do connect(); accept(); to establish a new channel.
    
    See Connnection for more details on this interface.

    """
    def __init__(self,
                 access_task: Task,
                 access_ram: RAM,
                 access_epoller: Epoller,
                 access_address: WrittenPointer[Sockaddr],
                 task: Task,
                 ram: RAM,
                 listening_fd: AsyncFileDescriptor,
    ) -> None:
        self.access_task = access_task
        self.access_ram = access_ram
        self.access_epoller = access_epoller
        self.access_address = access_address
        self.task = task
        self.ram = ram
        self.listening_fd = listening_fd

    async def open_async_channel(self) -> t.Tuple[AsyncFileDescriptor, FileDescriptor]:
        access_sock = await AsyncFileDescriptor.make(
            self.access_epoller, self.access_ram,
            await self.access_task.socket(self.access_address.value.family, SOCK.STREAM|SOCK.NONBLOCK))
        await access_sock.connect(self.access_address)
        sock = await self.listening_fd.accept()
        return access_sock, sock

    async def open_async_channels(self, count: int) -> t.List[t.Tuple[AsyncFileDescriptor, FileDescriptor]]:
        return [await self.open_async_channel() for _ in range(count)]

    async def open_channel(self) -> t.Tuple[FileDescriptor, FileDescriptor]:
        access_sock = await self.access_task.socket(self.access_address.value.family, SOCK.STREAM)
        # TODO this connect should really be async
        # but, since we're just connecting to a unix socket, it's fine I guess.
        await access_sock.connect(self.access_address)
        sock = await self.listening_fd.accept()
        return access_sock, sock

    async def open_channels(self, count: int) -> t.List[t.Tuple[FileDescriptor, FileDescriptor]]:
        return [await self.open_channel() for _ in range(count)]

    async def prep_fd_transfer(self) -> t.Tuple[FileDescriptor, t.Callable[[Task, RAM, FileDescriptor], Connection]]:
        return self.listening_fd.handle, self.for_task_with_fd

    def for_task_with_fd(self, task: Task, ram: RAM, fd: FileDescriptor) -> ListeningConnection:
        return ListeningConnection(
            self.access_task,
            self.access_ram,
            self.access_epoller,
            self.access_address,
            task, ram, self.listening_fd.with_handle(fd),
        )

    def inherit(self, task: Task, ram: RAM) -> ListeningConnection:
        return self.for_task_with_fd(task, ram, self.listening_fd.handle.inherit(task))



from rsyscall import local_thread
from rsyscall.fcntl import O
from rsyscall.sched import CLONE
from rsyscall.tests.trio_test_case import TrioTestCase
import gc

class TestFS(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread

    async def test_fd_gc(self) -> None:
        "Verify that file descriptors actually get GC'd."
        gc.collect()
        await self.thr.task.run_fd_table_gc()
        devnull = await self.thr.ptr("/dev/null")
        first = int(await self.thr.task.open(devnull, O.RDONLY))
        for _ in range(5):
            child = await self.thr.clone(CLONE.FILES)
            for _ in range(50):
                await child.task.open(devnull, O.RDONLY)
        gc.collect()
        await self.thr.task.run_fd_table_gc()
        last = int(await self.thr.task.open(devnull, O.RDONLY))
        self.assertEqual(first, last)



# Generated by the "nixdeps" line in setup.py
import nixdeps
import pathlib
closure = nixdeps.PackageClosure(pathlib.Path("/nix/store/i0sc43hrfspfw5qv39q16zbbd77jpp4l-rsyscall"), [
    "/nix/store/fnkixi37wfz5nlyzwab2l51p29793a1m-libunistring-0.9.10",
    "/nix/store/ai6j2i00rik53akq8r5pi2nqjvrqi7ky-libidn2-2.3.0",
    "/nix/store/90illc73xfs933d06daq6d41njs8yh66-glibc-2.32-37",
    "/nix/store/i0sc43hrfspfw5qv39q16zbbd77jpp4l-rsyscall",
])



"`#include <sys/syscall.h>`"
from rsyscall._raw import ffi, lib # type: ignore
import enum

class SYS(enum.IntEnum):
    """The syscall number argument passed to the low-level `syscall` method and underlying instruction

    Passing one of these numbers is how a userspace program indicates to the kernel which
    syscall it wants to call.

    """
    accept4 = lib.SYS_accept4
    bind = lib.SYS_bind
    capget = lib.SYS_capget
    capset = lib.SYS_capset
    chdir = lib.SYS_chdir
    chroot = lib.SYS_chroot
    clone = lib.SYS_clone
    close = lib.SYS_close
    connect = lib.SYS_connect
    dup3 = lib.SYS_dup3
    epoll_create1 = lib.SYS_epoll_create1
    epoll_ctl = lib.SYS_epoll_ctl
    epoll_wait = lib.SYS_epoll_wait
    eventfd2 = lib.SYS_eventfd2
    execve = lib.SYS_execve
    execveat = lib.SYS_execveat
    exit = lib.SYS_exit
    faccessat = lib.SYS_faccessat
    fchdir = lib.SYS_fchdir
    fchmod = lib.SYS_fchmod
    fcntl = lib.SYS_fcntl
    fstat = lib.SYS_fstat
    ftruncate = lib.SYS_ftruncate
    getdents64 = lib.SYS_getdents64
    getgid = lib.SYS_getgid
    getpeername = lib.SYS_getpeername
    getpgid = lib.SYS_getpgid
    getpid = lib.SYS_getpid
    getpriority = lib.SYS_getpriority
    getsockname = lib.SYS_getsockname
    getsockopt = lib.SYS_getsockopt
    getuid = lib.SYS_getuid
    inotify_add_watch = lib.SYS_inotify_add_watch
    inotify_init1 = lib.SYS_inotify_init1
    inotify_rm_watch = lib.SYS_inotify_rm_watch
    ioctl = lib.SYS_ioctl
    kill = lib.SYS_kill
    linkat = lib.SYS_linkat
    listen = lib.SYS_listen
    lseek = lib.SYS_lseek
    memfd_create = lib.SYS_memfd_create
    mkdirat = lib.SYS_mkdirat
    mmap = lib.SYS_mmap
    mount = lib.SYS_mount
    munmap = lib.SYS_munmap
    openat = lib.SYS_openat
    pipe2 = lib.SYS_pipe2
    prctl = lib.SYS_prctl
    pread64 = lib.SYS_pread64
    preadv2 = lib.SYS_preadv2
    prlimit64 = lib.SYS_prlimit64
    pwrite64 = lib.SYS_pwrite64
    pwritev2 = lib.SYS_pwritev2
    read = lib.SYS_read
    readlinkat = lib.SYS_readlinkat
    recvfrom = lib.SYS_recvfrom
    recvmsg = lib.SYS_recvmsg
    renameat2 = lib.SYS_renameat2
    rt_sigaction = lib.SYS_rt_sigaction
    rt_sigprocmask = lib.SYS_rt_sigprocmask
    sched_setaffinity = lib.SYS_sched_setaffinity
    sched_getaffinity = lib.SYS_sched_getaffinity
    sendmsg = lib.SYS_sendmsg
    sendto = lib.SYS_sendto
    set_robust_list = lib.SYS_set_robust_list
    set_tid_address = lib.SYS_set_tid_address
    setns = lib.SYS_setns
    setpgid = lib.SYS_setpgid
    setpriority = lib.SYS_setpriority
    setsid = lib.SYS_setsid
    setsockopt = lib.SYS_setsockopt
    shutdown = lib.SYS_shutdown
    signalfd4 = lib.SYS_signalfd4
    socket = lib.SYS_socket
    socketpair = lib.SYS_socketpair
    symlinkat = lib.SYS_symlinkat
    timerfd_create = lib.SYS_timerfd_create
    timerfd_gettime = lib.SYS_timerfd_gettime
    timerfd_settime = lib.SYS_timerfd_settime
    umount2 = lib.SYS_umount2
    unlinkat = lib.SYS_unlinkat
    unshare = lib.SYS_unshare
    waitid = lib.SYS_waitid
    write = lib.SYS_write



"Just the outcome library, with a little more typing"
from outcome import Value, Error
import outcome
import typing as t

__all__ = [
    'Outcome',
    'Value',
    'Error',
]

T = t.TypeVar('T')
class Outcome(t.Generic[T], outcome.Outcome):
    pass



"""A stub which can be launched by unrelated programs, which connects back to us

We create a StubServer, which listens on a Unix socket.

Then we arrange for other programs to launch rsyscall-unix-stub,
under an arbitrary executable name and with arbitrary arguments,
either with the RSYSCALL_UNIX_STUB_SOCK environment variable set or
through a wrapper shell script which sets RSYSCALL_UNIX_STUB_SOCK.

rsyscall-unix-stub will connect back to us over the Unix socket, and
we can call accept() to get a Thread which controls the
rsyscall-unix-stub, along with the command line arguments passed to
rsyscall-unix-stub, including the executable name.

At no point does the other program need to know that it is launching
the stub, nor does it have to be related to us in any way, other than
sharing enough of the mount table for the stub to connect back to the
Unix socket.

We can use these stubs to "mock" other programs. For example, we
could mock sendmail by putting a wrapper shell script named "sendmail"
on the PATH of some other program; then that program will run the stub
with stdin pointing to some email message and some command line
arguments. We call accept() and receive those arguments and a Thread
with that stdin, and we can do whatever we want with it.

We can make an analogy to the concept of a callback. In many
programming languages, we can pass a callback to other modules, and
when we're called we have complete control and we can choose what to
do and what to return. The stub is our way to pass a callback to
another program.

"""
from __future__ import annotations
import typing as t
import os
import rsyscall.near.types as near
import rsyscall.far as far
import rsyscall.handle as handle
from rsyscall.thread import Thread
from rsyscall.tasks.connection import SyscallConnection
from rsyscall.loader import NativeLoader
import trio
from dataclasses import dataclass
import logging
import rsyscall.memory.allocator as memory
from rsyscall.memory.ram import RAM
import rsyscall.nix as nix
from rsyscall.epoller import Epoller, AsyncFileDescriptor, AsyncReadBuffer
from rsyscall.monitor import ChildProcessMonitor
from rsyscall.command import Command
from rsyscall.memory.socket_transport import SocketMemoryTransport

import rsyscall.struct
from rsyscall.environ import Environment
from rsyscall.handle import WrittenPointer, FileDescriptor, Task

from rsyscall.path import Path
from rsyscall.sched import CLONE
from rsyscall.sys.socket import SOCK, AF, SendmsgFlags, SendMsghdr, CmsgList, CmsgSCMRights
from rsyscall.sys.mman import MFD
from rsyscall.sys.uio import IovecList
from rsyscall.sys.un import SockaddrUn
from rsyscall.signal import SIG, Sigset, SignalBlock
from rsyscall.fcntl import O

__all__ = [
    "StubServer",
]

logger = logging.getLogger(__name__)

@dataclass
class StubServer:
    "A server which can be connected back to by rsyscall-unix-stub."
    listening_sock: AsyncFileDescriptor
    thread: Thread

    @classmethod
    async def listen_on(cls, thread: Thread, path: Path) -> StubServer:
        "Start listening on the passed-in path for stub connections."
        sockfd = await thread.make_afd(await thread.socket(AF.UNIX, SOCK.STREAM|SOCK.NONBLOCK))
        addr = await thread.ram.ptr(await SockaddrUn.from_path(thread, path))
        await sockfd.handle.bind(addr)
        await sockfd.handle.listen(10)
        return StubServer(sockfd, thread)

    @classmethod
    async def make(cls, thread: Thread, dir: Path, name: str) -> StubServer:
        "In the passed-in dir, make a listening stub server and an executable to connect to it."
        import rsyscall._nixdeps.librsyscall
        rsyscall_path = await nix.deploy(thread, rsyscall._nixdeps.librsyscall.closure)
        stub_path = rsyscall_path/"libexec"/"rsyscall"/"rsyscall-unix-stub"
        sock_path = dir/f'{name}.sock'
        server = await StubServer.listen_on(thread, sock_path)
        # there's no POSIX sh way to set $0, so we'll pass $0 as $1, $1 as $2, etc.
        # $0 will be the stub executable, so we'll need to drop $0 in StubServer.
        wrapper = """#!/bin/sh
RSYSCALL_UNIX_STUB_SOCK_PATH={sock} exec {bin} "$0" "$@"
""".format(sock=os.fsdecode(sock_path), bin=os.fsdecode(stub_path))
        await thread.spit(dir/name, wrapper, mode=0o755)
        return server

    async def accept(self, thread: Thread=None) -> t.Tuple[t.List[str], Thread]:
        """Accept new connection and bootstrap over it; returns the stub's argv and the new thread

        The optional argument "thread" allows choosing what thread handles the bootstrap;
        by default we'll use the thread embedded in the StubServer.

        """
        if thread is None:
            thread = self.thread
        conn = await self.listening_sock.accept()
        argv, new_thread = await _setup_stub(thread, conn)
        # have to drop first argument, which is the unix_stub executable; see StubServer.make
        return argv[1:], new_thread

async def _setup_stub(
        thread: Thread,
        bootstrap_sock: FileDescriptor,
) -> t.Tuple[t.List[str], Thread]:
    "Setup a stub thread"
    [(access_syscall_sock, passed_syscall_sock),
     (access_data_sock, passed_data_sock)] = await thread.open_async_channels(2)
    # memfd for setting up the futex
    futex_memfd = await thread.task.memfd_create(
        await thread.ram.ptr(Path("child_robust_futex_list")))
    # send the fds to the new process
    connection_fd, make_connection = await thread.connection.prep_fd_transfer()
    async def sendmsg_op(sem: RAM) -> WrittenPointer[SendMsghdr]:
        iovec = await sem.ptr(IovecList([await sem.malloc(bytes, 1)]))
        cmsgs = await sem.ptr(CmsgList([CmsgSCMRights([
            passed_syscall_sock, passed_data_sock, futex_memfd, connection_fd])]))
        return await sem.ptr(SendMsghdr(None, iovec, cmsgs))
    _, [] = await bootstrap_sock.sendmsg(await thread.ram.perform_batch(sendmsg_op), SendmsgFlags.NONE)
    # close our reference to fds that only the new process needs
    await passed_syscall_sock.invalidate()
    await passed_data_sock.invalidate()
    # close the socketpair
    await bootstrap_sock.invalidate()
    #### read describe to get all the information we need from the new process
    describe_buf = AsyncReadBuffer(access_data_sock)
    describe_struct = await describe_buf.read_cffi('struct rsyscall_unix_stub')
    argv_raw = await describe_buf.read_length_prefixed_array(describe_struct.argc)
    argv = [os.fsdecode(arg) for arg in argv_raw]
    environ = await describe_buf.read_envp(describe_struct.envp_count)
    #### build the new task
    pid = describe_struct.pid
    fd_table = handle.FDTable(pid)
    address_space = far.AddressSpace(pid)
    # we assume pid namespace is shared
    pidns = thread.task.pidns
    process = near.Process(pid)
    # we assume net namespace is shared - that's dubious...
    # we should make it possible to control the namespace sharing more, hmm.
    # TODO maybe the describe should contain the net namespace number? and we can store our own as well?
    # then we can automatically do it right
    base_task = Task(process, fd_table, address_space, pidns)
    remote_syscall_fd = base_task.make_fd_handle(near.FileDescriptor(describe_struct.syscall_fd))
    base_task.sysif = SyscallConnection(
        logger.getChild(str(process)),
        access_syscall_sock, access_syscall_sock,
        remote_syscall_fd, remote_syscall_fd,
    )
    allocator = memory.AllocatorClient.make_allocator(base_task)
    base_task.sigmask = Sigset({SIG(bit) for bit in rsyscall.struct.bits(describe_struct.sigmask)})
    ram = RAM(base_task,
              SocketMemoryTransport(access_data_sock,
                                    base_task.make_fd_handle(near.FileDescriptor(describe_struct.data_fd))),
              allocator)
    # TODO I think I can maybe elide creating this epollcenter and instead inherit it or share it, maybe?
    # I guess I need to write out the set too in describe
    epoller = await Epoller.make_root(ram, base_task)
    child_monitor = await ChildProcessMonitor.make(ram, base_task, epoller)
    connection = make_connection(base_task, ram,
                                 base_task.make_fd_handle(near.FileDescriptor(describe_struct.connecting_fd)))
    new_thread = Thread(
        task=base_task,
        ram=ram,
        connection=connection,
        loader=NativeLoader.make_from_symbols(base_task, describe_struct.symbols),
        epoller=epoller,
        child_monitor=child_monitor,
        environ=Environment.make_from_environ(base_task, ram, environ),
        stdin=base_task.make_fd_handle(near.FileDescriptor(0)),
        stdout=base_task.make_fd_handle(near.FileDescriptor(1)),
        stderr=base_task.make_fd_handle(near.FileDescriptor(2)),
    )
    #### TODO set up futex I guess
    remote_futex_memfd = near.FileDescriptor(describe_struct.futex_memfd)
    return argv, new_thread



from __future__ import annotations
import email
import os
import abc
import trio
import socket
from rsyscall.trio_test_case import TrioTestCase
from rsyscall.thread import Thread
from rsyscall.handle import FileDescriptor, Path
from rsyscall.command import Command
from dataclasses import dataclass
from rsyscall.inotify_watch import Inotify
from rsyscall.mktemp import update_symlink

from rsyscall.netinet.in_ import SockaddrIn
from rsyscall.sys.socket import AF, SOCK, Sockbuf
from rsyscall.fcntl import O
from rsyscall.unistd import SEEK
from rsyscall.sys.un import SockaddrUn
from rsyscall.sys.inotify import IN
from rsyscall.sys.memfd import MFD

class Dovecot:
    pass

class MailLocation:
    @abc.abstractmethod
    def spec(self) -> str: ...

@dataclass
class Maildir(MailLocation):
    path: Path
    @staticmethod
    async def make(thr: Thread, path: Path) -> Maildir:
        self = Maildir(path)
        await thr.mkdir(self.path)
        await thr.mkdir(self.new())
        return self

    def spec(self) -> str:
        return "maildir:" + os.fsdecode(self.path)

    def new(self) -> Path:
        return self.path/'new'

async def start_dovecot(nursery, thread: Thread, path: Path,
                        lmtp_listener: FileDescriptor, mail: MailLocation) -> Dovecot:
    dovecot = await thread.environ.which("dovecot")
    doveadm = await thread.environ.which("doveadm")
    s6_ipcserverd = await thread.environ.which("s6-ipcserverd")
    config = """
protocols =
log_path = /dev/stderr
mail_debug = yes
auth_debug = yes
service anvil {
  chroot =
}
service stats {
  chroot =
}
userdb {
  driver = static
  args = allow_all_users=yes
}
passdb {
  driver = passwd-file
  args = /dev/null
}
"""
    config += "base_dir = " + os.fsdecode(await thread.mkdir(path/"base")) + "\n"
    config += "state_dir = " + os.fsdecode(await thread.mkdir(path/"state")) + "\n"
    # unfortunately, dovecot requires names for these configuration parameters, and
    # doesn't accept ids. would be a nice patch to upstream...
    # TODO get these with id -n{u,g} I guess?
    username = "sbaugh"
    groupname = "sbaugh"
    config += f"default_login_user = {username}\n"
    config += f"default_internal_user = {username}\n"
    config += f"default_internal_group = {groupname}\n"
    # all mail we get from the socket goes to a single destination: this maildir
    config += f"mail_location = {mail.spec()}\n"

    config_path = await thread.spit(path/"dovecot.conf", config)
    # start dovecot
    dovecot_thread = await thread.clone()
    dovecot_child = await dovecot_thread.exec(dovecot.args('-F', '-c', config_path))
    nursery.start_soon(dovecot_child.check)

    # start lmtp server
    lmtp_thread = await thread.clone()
    lmtp_listener = lmtp_listener.move(lmtp_thread.task)
    await lmtp_thread.unshare_files(going_to_exec=True)
    await lmtp_thread.stdin.replace_with(lmtp_listener)
    lmtp_child = await lmtp_thread.exec(s6_ipcserverd.args(
        doveadm.executable_path, '-c', config_path, 'exec', 'lmtp'))
    nursery.start_soon(lmtp_child.check)
    return Dovecot()

@dataclass
class Smtpd:
    lmtp_socket_path: Path
    lmtp_listener: FileDescriptor
    config_file: Path

async def start_smtpd(nursery, thread: Thread, path: Path,
                      smtp_listener: FileDescriptor) -> Smtpd:
    smtpd = await thread.environ.which("smtpd")
    smtpd_thread = await thread.clone()
    smtp_listener = smtp_listener.move(smtpd_thread.task)
    await smtpd_thread.unshare_files()

    config = ""
    config += 'listen on socket path "' + os.fsdecode(path/"smtpd.sock") + '"\n'
    config += "table aliases file:" + os.fsdecode(await thread.spit(path/"aliases", "")) + "\n"
    config += 'queue path "' + os.fsdecode(await thread.mkdir(path/"spool", mode=0o711)) + '"\n'
    config += 'path chroot "' + os.fsdecode(await thread.mkdir(path/"empty")) + '"\n'
    config += "listen on localhost inherit " + str(await smtp_listener.as_argument()) + '\n'

    # bind a socket in the parent
    lmtp_socket = await thread.task.socket(AF.UNIX, SOCK.STREAM)
    lmtp_socket_path = path/"lmtp.sock"
    await lmtp_socket.bind(await thread.ram.ptr(SockaddrUn(os.fsencode(lmtp_socket_path))))
    await lmtp_socket.listen(10)
    config += 'action "local" lmtp "' + os.fsdecode(lmtp_socket_path) + '" user root\n'
    # all mail is delivered to this single socket
    # TODO actually dispatch correctly: we need one socket per username we accept
    config += 'match from any auth for local action "local"\n'

    config += 'action "relay" relay\n'
    config += 'match from local auth for any action "relay"\n'

    # smtpd has a lot of asserts that it is running as root, even
    # though we could arrange things so that it doesn't actually need
    # root. we'll put it in a user namespace so it gets its wish, and
    # run everything under "root", without privsep. "root" of course
    # is mapped to an unpriv user, so this is close to the same
    # security guarantee. we don't get separation between the main
    # user and the queue user, though... alas.
    await smtpd_thread.unshare_user(in_namespace_uid=0, in_namespace_gid=0)
    config += "queue user root\n"
    config += "queue group root\n"
    config += "smtp user root\n"

    config_path = await thread.spit(path/"smtpd.config", config)
    child = await smtpd_thread.exec(smtpd.args("-v", "-d", "-f", config_path))
    nursery.start_soon(child.check)

    return Smtpd(
        lmtp_socket_path=lmtp_socket_path,
        lmtp_listener=lmtp_socket,
        config_file=config_path,
    )

import rsyscall.tasks.local as local
class TestMail(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thread = local.thread
        self.tmpdir = await self.thread.mkdtemp("test_mail")
        self.path = self.tmpdir.path
        await update_symlink(self.thread, await self.thread.ram.ptr(self.tmpdir.parent/"test_mail.current"), self.path)
        smtp_sock = await self.thread.task.socket(AF.INET, SOCK.STREAM)
        await smtp_sock.bind(await self.thread.ram.ptr(SockaddrIn(3000, "127.0.0.1")))
        await smtp_sock.listen(10)
        self.smtpd = await start_smtpd(self.nursery, self.thread, await self.thread.mkdir(self.path/"smtpd"), smtp_sock)
        self.maildir = await Maildir.make(self.thread, self.path/"mail")
        self.dovecot = await start_dovecot(self.nursery, self.thread, await self.thread.mkdir(self.path/"dovecot"),
                                           self.smtpd.lmtp_listener, self.maildir)
        smtpctl = await self.thread.environ.which("smtpctl")
        self.sendmail = Command(smtpctl.executable_path, [b'sendmail'], {'SMTPD_CONFIG_FILE': self.smtpd.config_file})
        self.inty = await Inotify.make(self.thread)


    async def asyncTearDown(self) -> None:
        await self.tmpdir.cleanup()

    async def send_email(self, from_: str, to: str, subject: str, msg: str) -> None:
        thread = await self.thread.clone()
        await thread.unshare_files()
        fd = await thread.task.memfd_create(await thread.ram.ptr(Path('message')))
        msg = f'From: {from_}\nSubject: {subject}\nTo: {to}\n\n' + msg
        await thread.spit(fd, msg)
        await fd.lseek(0, SEEK.SET)
        await thread.stdin.replace_with(fd)
        child = await thread.exec(self.sendmail.args('-t'))
        await child.check()

    async def test_sendmail(self) -> None:
        watch = await self.inty.add(self.maildir.new(), IN.MOVED_TO)
        # TODO sigh, opensmtpd isn't validating the From field...
        from_ = 'from@localhost'
        to = 'sbaugh@localhost'
        subject = 'Subjective'
        msg = 'Hello me!\n'
        await self.send_email(from_, to, subject, msg)
        event = await watch.wait_until_event(IN.MOVED_TO)
        if event.name is None:
            raise Exception("event has no name??")
        mailfd = await self.thread.task.open(await self.thread.ram.ptr(self.maildir.new()/event.name), O.RDONLY)
        data = await self.thread.read_to_eof(mailfd)
        message = email.message_from_bytes(data)
        self.assertEqual(from_,  message['From'])
        self.assertEqual(to,  message['To'])
        self.assertEqual(subject, message['Subject'])
        self.assertEqual(msg, message.get_payload())

    # So I need to set up proper DNS stuff.
    # Which... I can do by running my own DNS server :)
    # aha, okay, so I could have a DNS server which,
    # automatically forwards the requests to the NS server in the record
    async def test_mail_tester(self) -> None:
        from_ = 'sbaugh@catern.com'
        to = 'test-2ux4p@mail-tester.com'
        subject = 'Subjective'
        msg = 'Hello me!\n'
        await self.send_email(from_, to, subject, msg)
        await trio.sleep(9999)

if __name__ == "__main__":
    import unittest
    unittest.main()



"`#include <sys/epoll.h>`"
from __future__ import annotations
from rsyscall._raw import lib, ffi # type: ignore
from rsyscall.struct import Struct, Serializable
import enum
import typing as t
from dataclasses import dataclass

class EpollFlag(enum.IntFlag):
    NONE = 0
    CLOEXEC = lib.EPOLL_CLOEXEC

class EPOLL_CTL(enum.IntEnum):
    ADD = lib.EPOLL_CTL_ADD
    MOD = lib.EPOLL_CTL_MOD
    DEL = lib.EPOLL_CTL_DEL

class EPOLL(enum.IntFlag):
    NONE = 0
    IN = lib.EPOLLIN
    OUT = lib.EPOLLOUT
    RDHUP = lib.EPOLLRDHUP # type: ignore
    PRI = lib.EPOLLPRI
    ERR = lib.EPOLLERR
    HUP = lib.EPOLLHUP
    # options
    ET = lib.EPOLLET

    def __iter__(self) -> t.Iterator[EPOLL]:
        for flag in EPOLL:
            if self & flag:
                yield flag

@dataclass
class EpollEvent(Struct):
    data: int
    events: EPOLL

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct epoll_event const*', (self.events, (self.data,)))))

    T = t.TypeVar('T', bound='EpollEvent')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct epoll_event*', ffi.from_buffer(data))
        return cls(struct.data.u64, EPOLL(struct.events))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct epoll_event')

class EpollEventList(t.List[EpollEvent], Serializable):
    def to_bytes(self) -> bytes:
        ret = b""
        for ent in self:
            ret += ent.to_bytes()
        return ret

    T = t.TypeVar('T', bound='EpollEventList')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        entries = []
        while len(data) > 0:
            ent = EpollEvent.from_bytes(data)
            entries.append(ent)
            data = data[EpollEvent.sizeof():]
        return cls(entries)

#### Classes ####
from rsyscall.handle.fd import BaseFileDescriptor, FileDescriptorTask
from rsyscall.handle.pointer import Pointer, WrittenPointer, ReadablePointer

T_fd = t.TypeVar('T_fd', bound='EpollFileDescriptor')
class EpollFileDescriptor(BaseFileDescriptor):
    async def epoll_wait(self, events: Pointer[EpollEventList],
                         timeout: int) -> t.Tuple[ReadablePointer[EpollEventList], Pointer]:
        self._validate()
        with events.borrow(self.task):
            maxevents = events.size()//EpollEvent.sizeof()
            num = await _epoll_wait(self.task.sysif, self.near, events.near, maxevents, timeout)
            valid_size = num * EpollEvent.sizeof()
            return events.readable_split(valid_size)

    async def epoll_ctl(self, op: EPOLL_CTL, fd: BaseFileDescriptor,
                        event: t.Optional[WrittenPointer[EpollEvent]]=None) -> None:
        self._validate()
        with fd.borrow(self.task):
            if event is not None:
                event.check_address_space(self.task)
            return (await _epoll_ctl(self.task.sysif, self.near, op, fd.near, event.near if event else None))

class EpollTask(FileDescriptorTask[T_fd]):
    async def epoll_create(self, flags: EpollFlag=EpollFlag.NONE) -> T_fd:
        return self.make_fd_handle(await _epoll_create(self.sysif, flags|EpollFlag.CLOEXEC))

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _epoll_create(sysif: SyscallInterface, flags: EpollFlag) -> near.FileDescriptor:
    return near.FileDescriptor(await sysif.syscall(SYS.epoll_create1, flags))

async def _epoll_ctl(sysif: SyscallInterface, epfd: near.FileDescriptor, op: EPOLL_CTL,
                     fd: near.FileDescriptor, event: t.Optional[near.Address]=None) -> None:
    if event is None:
        event = 0 # type: ignore
    await sysif.syscall(SYS.epoll_ctl, epfd, op, fd, event)

async def _epoll_wait(sysif: SyscallInterface, epfd: near.FileDescriptor,
                      events: near.Address, maxevents: int, timeout: int) -> int:
    return (await sysif.syscall(SYS.epoll_wait, epfd, events, maxevents, timeout))


#### Tests ####
from unittest import TestCase
class TestEpoll(TestCase):
    def test_epoll_event_list(self) -> None:
        initial = EpollEventList([EpollEvent(42, EPOLL.IN|EPOLL.PRI)])
        output = EpollEventList.from_bytes(initial.to_bytes())
        self.assertEqual(initial, output)



import unittest
from arepl import *
import typing as t

T = t.TypeVar('T')
def await_pure(awaitable: t.Awaitable[T]) -> T:
    iterable = awaitable.__await__()
    try:
        next(iterable)
    except StopIteration as e:
        return e.value
    else:
        raise Exception("this awaitable actually is impure! it yields!")

class TestPure(unittest.TestCase):
    def test_add(self) -> None:
        async def test() -> None:
            repl = PureREPL({})
            async def eval(line: str) -> t.Any:
                result = await repl.add_line(line + '\n')
                if isinstance(result, ExpressionResult):
                    return result.value
                else:
                    raise Exception("unexpected", result)
            self.assertEqual(await eval('1'), 1)
            self.assertEqual(await eval('1+1'), 2)
            await repl.add_line('foo = 1\n')
            self.assertEqual(await eval('foo*4'), 4)
        await_pure(test())



"`#include <netinet/in.h>`"
from rsyscall._raw import ffi, lib # type: ignore
from rsyscall.sys.socket import Sockaddr, AF, _register_sockaddr
import ipaddress
import socket
import typing as t
from dataclasses import dataclass

__all__ = [
    'SockaddrIn',
    'SockaddrIn6',
]

@dataclass(frozen=True)
class SockaddrIn(Sockaddr):
    """Representation of struct sockaddr_in

    The port is in totally-normal host byte order, even though with the real
    struct sockaddr_in, the port (and address) would be in network byte
    order. That's just an encoding quirk of C, not something we want to copy.

    """
    port: int
    addr: ipaddress.IPv4Address
    family = AF.INET
    def __init__(self, port: int, addr: t.Union[str, int, ipaddress.IPv4Address]) -> None:
        # the dataclass is frozen so we have to use __setattr__
        object.__setattr__(self, 'port', port)
        object.__setattr__(self, 'addr', ipaddress.IPv4Address(addr))

    def to_bytes(self) -> bytes:
        addr = ffi.new('struct sockaddr_in*', (AF.INET, socket.htons(self.port), (socket.htonl(int(self.addr)),)))
        return bytes(ffi.buffer(addr))

    T = t.TypeVar('T', bound='SockaddrIn')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        if len(data) < cls.sizeof():
            raise Exception("data too small", data)
        struct = ffi.cast('struct sockaddr_in*', ffi.from_buffer(data))
        cls.check_family(AF(struct.sin_family))
        return cls(socket.ntohs(struct.sin_port), socket.ntohl(struct.sin_addr.s_addr))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct sockaddr_in')

    def addr_as_string(self) -> str:
        "Returns the addr portion of this address in 127.0.0.1 form"
        return str(self.addr)

    def __str__(self) -> str:
        return f"SockaddrIn({self.addr_as_string()}:{self.port})"

    def __repr__(self) -> str:
        return str(self)
_register_sockaddr(SockaddrIn)


class SockaddrIn6(Sockaddr):
    "Representation of struct sockaddr_in6"
    family = AF.INET6
    def __init__(self, port: int, addr: t.Union[str, int, ipaddress.IPv6Address],
                 flowinfo: int=0, scope_id: int=0) -> None:
        # these are in host byte order, of course
        self.port = port
        self.addr = ipaddress.IPv6Address(addr)
        self.flowinfo = flowinfo
        self.scope_id = scope_id

    def to_bytes(self) -> bytes:
        struct = ffi.new('struct sockaddr_in6*',
                         (AF.INET6, socket.htons(self.port), self.flowinfo, (b'',), self.scope_id))
        struct.sin6_addr.s6_addr = self.addr.packed
        return bytes(ffi.buffer(struct))

    T = t.TypeVar('T', bound='SockaddrIn6')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        if len(data) < cls.sizeof():
            raise Exception("data too small", data)
        struct = ffi.cast('struct sockaddr_in6*', ffi.from_buffer(data))
        cls.check_family(AF(struct.sin6_family))
        return cls(socket.ntohs(struct.sin6_port), ipaddress.IPv6Address(bytes(struct.sin6_addr.s6_addr)),
                   struct.sin6_flowinfo, struct.sin6_scope_id)

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct sockaddr_in6')

    def __str__(self) -> str:
        return f"SockaddrIn6({self.addr}:{self.port})"

    def __repr__(self) -> str:
        return str(self)
_register_sockaddr(SockaddrIn6)


#### Tests ####
from unittest import TestCase
class TestIn(TestCase):
    def test_sockaddrin(self) -> None:
        initial = SockaddrIn(42, "127.0.0.1")
        output = SockaddrIn.from_bytes(initial.to_bytes())
        self.assertEqual(initial.port, output.port)
        self.assertEqual(initial.addr, output.addr)
        from rsyscall.sys.socket import SockaddrStorage
        out = SockaddrStorage.from_bytes(initial.to_bytes()).parse()
        self.assertEqual(initial.port, output.port)
        self.assertEqual(initial.addr, output.addr)

    def test_sockaddrIn6(self) -> None:
        initial = SockaddrIn6(42, "34:12::19:9a")
        output = SockaddrIn6.from_bytes(initial.to_bytes())
        self.assertEqual(initial.port, output.port)
        self.assertEqual(initial.addr, output.addr)
        from rsyscall.sys.socket import SockaddrStorage
        out = SockaddrStorage.from_bytes(initial.to_bytes()).parse()
        self.assertEqual(initial.port, output.port)
        self.assertEqual(initial.addr, output.addr)
        



"Functions for making temporary directories."
import random
import string
from rsyscall.thread import Thread
from rsyscall.path import Path
from rsyscall.handle import WrittenPointer
import os
import typing as t

def random_string(k: int=8) -> str:
    "Return a random string - useful for making files that don't conflict with others."
    return ''.join(random.choices(string.ascii_letters + string.digits, k=k))

async def update_symlink(thr: Thread, path: WrittenPointer[Path],
                         target: t.Union[str, os.PathLike]) -> WrittenPointer[Path]:
    "Atomically update this path to contain a symlink pointing at this target."
    tmpname = path.value.name + ".updating." + random_string(k=8)
    tmppath = await thr.ram.ptr(path.value.parent/tmpname)
    await thr.task.symlink(await thr.ram.ptr(target), tmppath)
    await thr.task.rename(tmppath, path)
    return path

async def mkdtemp(thr: Thread, prefix: str="mkdtemp") -> 'TemporaryDirectory':
    "Make a temporary directory in thr.environ.tmpdir."
    parent = thr.environ.tmpdir
    name = prefix+"."+random_string(k=8)
    await thr.task.mkdir(await thr.ram.ptr(parent/name), 0o700)
    return TemporaryDirectory(thr, parent, name)

class TemporaryDirectory(Path):
    "A temporary directory we've created and are responsible for cleaning up."
    def __init__(self, thr: Thread, parent: Path, name: str) -> None:
        "Don't directly instantiate, use rsyscall.mktemp.mkdtemp to create this class."
        self.thr = thr
        super().__init__(parent, name)

    async def cleanup(self) -> None:
        """Delete this temporary directory and everything inside it.

        We do this cleanup by execing sh; that's the cheapest way to do it.  We have to
        chmod -R +w the directory before we rm -rf it, because the directory might contain
        files without the writable bit set, which would prevent us from deleting it.

        """
        # TODO would be nice if not sharing the fs information gave us a cap to chdir
        cleanup = await self.thr.clone()
        await cleanup.task.chdir(await cleanup.ram.ptr(self.parent))
        child = await cleanup.exec(self.thr.environ.sh.args(
            '-c', f"chmod -R +w -- {self.name} && rm -rf -- {self.name}"))
        await child.check()

    async def __aenter__(self) -> Path:
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.cleanup()



from __future__ import annotations
import unittest

from rsyscall.tests.trio_test_case import TrioTestCase

from rsyscall import local_thread

from rsyscall.thread import do_cloexec_except

from rsyscall.tests.utils import do_async_things
from rsyscall.fcntl import O
from rsyscall.unistd import Pipe
from rsyscall.sched import CLONE

class TestMisc(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.local = local_thread
        self.thr = await self.local.clone()

    async def asyncTearDown(self) -> None:
        await self.thr.exit(0)

    async def test_do_cloexec_except(self) -> None:
        pipe = await (await self.thr.task.pipe(await self.thr.ram.malloc(Pipe))).read()
        close_set = set([fd.near for fd in self.thr.task.fd_handles])
        close_set.remove(pipe.read.near)
        await do_cloexec_except(self.thr, close_set)

        data = await self.thr.ram.ptr(b"foo")
        with self.assertRaises(OSError):
            # this side was closed due to being cloexec
            await pipe.read.read(data)
        with self.assertRaises(BrokenPipeError):
            # this side is still open, but gets EPIPE
            await pipe.write.write(data)



"Interfaces and basic functionality for our serialization framework."
from __future__ import annotations
import abc
import os
import typing as t
import struct
from dataclasses import dataclass

T = t.TypeVar('T')
class Serializer(t.Generic[T]):
    """An object encapsulating the information required to serialize or deserialize some value of some type.

    This may only work for one specific value of that type; for example, to
    deserialize a type containing pointers (encoded as addresses), we must have
    a serializer that knows about those addresses and can return the
    corresponding Python object Pointers.

    """
    def to_bytes(self, val: T) -> bytes:
        "Serialize this value `val` to bytes"
        raise NotImplementedError("to_bytes not implemented on", type(self))

    def from_bytes(self, data: bytes) -> T:
        "Deserialize a value from `data`"
        raise NotImplementedError("from_bytes not implemented on", type(self))

T_has_serializer = t.TypeVar('T_has_serializer', bound='HasSerializer')
class HasSerializer:
    "Something which can return a serializer for itself"
    @abc.abstractmethod
    def get_self_serializer(self: T_has_serializer, task) -> Serializer[T_has_serializer]:
        """Return a serializer for this value

        Note that, as discussed in the Serializer docstring, the serializer
        returned may only work for the specific value that this method was
        called on, not any other values of the same type.

        """
        pass

T_fixed_serializer = t.TypeVar('T_fixed_serializer', bound='FixedSerializer')
class FixedSerializer(HasSerializer):
    "Something which, if we know its class, can be serialized and deserialized"
    @classmethod
    @abc.abstractmethod
    def get_serializer(cls: t.Type[T_fixed_serializer], task) -> Serializer[T_fixed_serializer]:
        "Return a Serializer for this class"
        pass

    def get_self_serializer(self: T_fixed_serializer, task) -> Serializer[T_fixed_serializer]:
        """Return a Serializer for this value

        Since this class is a FixedSerializer, we can return a serializer that is
        independent of any specific value by calling get_serializer. Which we do.

        """
        return type(self).get_serializer(task)

T_fixed_size = t.TypeVar('T_fixed_size', bound='FixedSize')
class FixedSize(FixedSerializer):
    """Something which, if we know its class, can be serialized and deserialized to a fixed length bytestring

    These are fixed-size structures; for example, most C structs.
    """
    @classmethod
    @abc.abstractmethod
    def sizeof(cls) -> int:
        "Return the length of the bytestring that the serializer for this class will return"
        pass

class Serializable(FixedSerializer):
    "A helper class for FixedSerializer; the serialization methods are defined directly on the class"
    @abc.abstractmethod
    def to_bytes(self) -> bytes:
        """Directly serialize the value `self` as bytes

        We use this as the Serializer.to_bytes method through the magic of duck typing;
        using this function from the underlying class gives us a function which takes a
        value (which it calls `self`) and does serialization for that value.
        """
        pass

    T_serializable = t.TypeVar('T_serializable', bound='Serializable')
    @classmethod
    def from_bytes(cls: t.Type[T_serializable], data: bytes) -> T_serializable:
        "Return a value of type `cls` deserialized from `data`"
        raise NotImplementedError("from_bytes not implemented on", cls)
    @classmethod
    def get_serializer(cls: t.Type[T_serializable], task) -> Serializer[T_serializable]: # type: ignore
        """Return a "serializer" for this class - it's actually just this class itself.

        Yay, duck typing!
        """
        return cls # type: ignore

class Struct(Serializable, FixedSize):
    "A helper class for FixedSize; the serialization methods are defined directly on the class"
    pass

def bits(n: int, one_indexed: bool=True) -> t.Iterator[int]:
    "Yield the bit indices that are set in this integer"
    while n:
        b = n & (~n+1)
        yield (b.bit_length() - (0 if one_indexed else 1))
        n ^= b

# mypy is very upset with me for inheriting from int and overriding int's methods in an incompatible way
class Int32(Struct, int): # type: ignore
    "A 32-bit integer, as used by many syscalls"
    def to_bytes(self) -> bytes: # type: ignore
        return struct.pack('i', self)

    T = t.TypeVar('T', bound='Int32')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T: # type: ignore
        if len(data) < cls.sizeof():
            raise Exception("data too small", data)
        val, = struct.unpack_from('i', data)
        return cls(val)
        
    @classmethod
    def sizeof(cls) -> int:
        return struct.calcsize('i')

class Int64(Struct, int): # type: ignore
    "A 64-bit integer, as used by many syscalls"
    def to_bytes(self) -> bytes: # type: ignore
        return struct.pack('l', self)

    T = t.TypeVar('T', bound='Int64')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T: # type: ignore
        if len(data) < cls.sizeof():
            raise Exception("data too small", data)
        val, = struct.unpack_from('l', data)
        return cls(val)

    @classmethod
    def sizeof(cls) -> int:
        return struct.calcsize('l')

@dataclass
class StructList(t.Generic[T_fixed_size], HasSerializer):
    "A list of serializable, fixed-size structures"
    cls: t.Type[T_fixed_size]
    elems: t.List[T_fixed_size]

    def get_self_serializer(self, task) -> StructListSerializer[T_fixed_size]:
        return StructListSerializer(self.cls, self.cls.get_serializer(task))

@dataclass
class StructListSerializer(t.Generic[T_fixed_size], Serializer[StructList[T_fixed_size]]):
    "The serializer for a StructList"
    cls: t.Type[T_fixed_size]
    ser: Serializer[T_fixed_size]

    def to_bytes(self, val: StructList[T_fixed_size]) -> bytes:
        return b"".join(self.ser.to_bytes(ent) for ent in val.elems)

    def from_bytes(self, data: bytes) -> StructList[T_fixed_size]:
        entries = []
        while len(data) > 0:
            ent = self.ser.from_bytes(data)
            entries.append(ent)
            data = data[self.cls.sizeof():]
        return StructList(self.cls, entries)

def strpath_to_null_terminated_bytes(val: t.Union[str, os.PathLike]) -> bytes:
    return os.fsencode(val) + b'\0'

def string_from_null_terminated_bytes(data: bytes) -> str:
    try:
        nullidx = data.index(b'\0')
    except ValueError:
        return os.fsdecode(data)
    else:
        return os.fsdecode(data[0:nullidx])

T_pathlike = t.TypeVar('T_pathlike', bound=os.PathLike)
@dataclass
class PathLikeSerializer(Serializer[T_pathlike]):
    cls: t.Type[T_pathlike]

    @staticmethod
    def to_bytes(val: T_pathlike) -> bytes:
        return strpath_to_null_terminated_bytes(val)

    def from_bytes(self, data: bytes) -> T_pathlike:
        # We assume that any PathLike can be constructed by just passing a single string to its constructor.
        # That's not actually true - os.DirEntry, for example, can't be constructed directly in this way.
        # But, this is Python, so if it quacks, it ships.
        return self.cls(string_from_null_terminated_bytes(data)) # type: ignore

class StrSerializer(Serializer[str]):
    @staticmethod
    def to_bytes(val: str) -> bytes:
        return strpath_to_null_terminated_bytes(val)

    @staticmethod
    def from_bytes(data: bytes) -> str:
        return string_from_null_terminated_bytes(data)



import logging
from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
from rsyscall.sys.prctl import *
logger = logging.getLogger(__name__)

class TestUser(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.local = local_thread
        self.thr = await self.local.clone()

    async def asyncTearDown(self) -> None:
        await self.thr.exit(0)

    async def test_name(self) -> None:
        namep = await self.thr.task.prctl(PR.GET_NAME, await self.thr.malloc(str, 16))
        logger.info("My initial name is %s", await namep.read())
        newname = "newname"
        await self.thr.task.prctl(PR.SET_NAME, await self.thr.ptr(newname))
        namep = await self.thr.task.prctl(PR.GET_NAME, namep)
        self.assertEqual(newname, await namep.read())



from __future__ import annotations
import typing as t
import time
from rsyscall._raw import lib # type: ignore
from rsyscall._raw import ffi # type: ignore
import os
import abc
import trio
import rsyscall.handle as handle
import rsyscall.near
from rsyscall.trio_test_case import TrioTestCase
from rsyscall.thread import Thread, ChildThread
from rsyscall.command import Command
from rsyscall.handle import FileDescriptor, Path
from dataclasses import dataclass
from rsyscall.struct import Int32
from rsyscall.monitor import AsyncChildProcess

import rsyscall.tasks.local as local
from rsyscall.sys.capability import CAP, CapHeader, CapData
from rsyscall.sys.socket import AF, SOCK, SOL, Socketpair
from rsyscall.sys.prctl import PR, PR_CAP_AMBIENT
from rsyscall.fcntl import O
from rsyscall.sched import CLONE
from rsyscall.netinet.in_ import SockaddrIn
from rsyscall.netinet.ip import IP, IPPROTO
from rsyscall.linux.netlink import SockaddrNl, NETLINK
from rsyscall.linux.rtnetlink import RTMGRP
from rsyscall.net.if_ import Ifreq, IFF, TUNSETIFF, SIOC

import rsyscall.nix as nix

from rsyscall._nixdeps.miredo import closure as miredo_nixdep

@dataclass
class MiredoExecutables:
    run_client: Command
    privproc: Command

    @classmethod
    async def from_store(cls, store: nix.Store) -> MiredoExecutables:
        miredo_path = await store.realise(miredo_nixdep)
        return MiredoExecutables(
            run_client=Command(miredo_path/"libexec"/"miredo"/"miredo-run-client", ["miredo-run-client"], {}),
            privproc=Command(miredo_path/"libexec"/"miredo"/"miredo-privproc", ["miredo-privproc"], {}),
        )

@dataclass
class Miredo:
    # we could use setns instead of keeping a thread around inside the namespace.
    # that would certainly be more lightweight.
    # but, the hassle with setns is that it seems you must setns to
    # the owning userns before you can setns to the netns.
    # you can't just do an unshare(USER) to get caps then setns to wherever.
    # I don't get why this is the case, and I'm not sure it can't be worked around.
    # So, I'll just use a thread, which I do understand.
    # Hopefully we can get a more lightweight setns-based approach later?
    ns_thread: ChildThread
    privproc_child: AsyncChildProcess
    client_child: AsyncChildProcess

async def add_to_ambient_caps(thr: Thread, capset: t.Set[CAP]) -> None:
    hdr = await thr.ptr(CapHeader())
    data_ptr = await thr.task.capget(hdr_ptr, await thr.malloc(CapData))
    data = await data_ptr.read()
    data.inheritable.update(capset)
    await thr.task.capset(hdr_ptr, await data_ptr.write(data))
    for cap in capset:
        await thr.task.prctl(PR.CAP_AMBIENT, PR_CAP_AMBIENT.RAISE, cap)

async def set_miredo_sockopts(thread: Thread, fd: FileDescriptor) -> None:
    # set a bunch of sockopts
    one = await thread.ram.ptr(Int32(1))
    await fd.setsockopt(SOL.IP, IP.RECVERR, one)
    await fd.setsockopt(SOL.IP, IP.PKTINFO, one)
    await fd.setsockopt(SOL.IP, IP.MULTICAST_TTL, one)
    # hello fragments my old friend
    await fd.setsockopt(SOL.IP, IP.MTU_DISCOVER, await thread.ram.ptr(Int32(IP.PMTUDISC_DONT)))

async def make_tun(thread: Thread, name: str, reqsock: FileDescriptor) -> t.Tuple[FileDescriptor, int]:
    # open /dev/net/tun
    tun_fd = await thread.task.open(await thread.ptr(Path("/dev/net/tun")), O.RDWR)
    # register TUN interface name for this /dev/net/tun fd
    ifreq = await thread.ptr(Ifreq(name, flags=IFF.TUN))
    await tun_fd.ioctl(TUNSETIFF, ifreq)
    # use reqsock to look up the interface index of the TUN interface by name (reusing the previous Ifreq)
    await reqsock.ioctl(SIOC.GIFINDEX, ifreq)
    tun_index = (await ifreq.read()).ifindex
    return tun_fd, tun_index

async def unmount_everything_except(thread: Thread, path: Path) -> None:
    pass

async def start_miredo_internal(thread: Thread, miredo_exec: MiredoExecutables) -> Miredo:
    ### create socket outside network namespace that Miredo will use for internet access
    inet_sock = await thread.socket(AF.INET, SOCK.DGRAM)
    await inet_sock.bind(await thread.ptr(SockaddrIn(0, 0)))
    # set some miscellaneous additional sockopts that Miredo wants
    await set_miredo_sockopts(thread, inet_sock)
    ### create main network namespace thread
    ns_thread = await thread.clone(CLONE.NEWNET|CLONE.NEWUSER)
    ### create in-network-namespace raw INET6 socket which Miredo will use to relay pings
    icmp6_fd = await ns_thread.socket(AF.INET6, SOCK.RAW, IPPROTO.ICMPV6)
    ### create in-network-namespace socket which Miredo will use for unassociated Ifreq ioctls
    reqsock = await ns_thread.socket(AF.INET, SOCK.STREAM)
    ### create and set up the TUN interface
    tun_fd, tun_index = await make_tun(ns_thread, "miredo", reqsock)
    ### create socketpair which Miredo will use to communicate between privileged process and Teredo client
    privproc_pair = await ns_thread.socketpair(AF.UNIX, SOCK.STREAM)
    ### start up privileged process which manipulates the network setup in the namespace
    # privproc_thread is cloned from ns_thread, which was also created through clone; this nesting is fully supported
    privproc_thread = await ns_thread.clone()
    # preserve NET_ADMIN capability over exec so that privproc can manipulate the TUN interface
    # helper function used because manipulating Linux ambient capabilities is fairly verbose
    await add_to_ambient_caps(privproc_thread, {CAP.NET_ADMIN})
    # privproc expects to communicate with the main client over stdin and stdout
    privproc_side = privproc_thread.inherit_fd(privproc_pair.first)
    await privproc_side.dup2(privproc_thread.stdin)
    await privproc_side.dup2(privproc_thread.stdout)
    privproc_child = await privproc_thread.execve(miredo_exec.privproc.executable_path, [
        miredo_exec.privproc.executable_path, str(tun_index)
    ])
    ### start up Miredo client process which communicates over the internet to implement the tunnel
    # the client process doesn't need to be in the same network namespace, since it is passed all
    # the resources it needs as fds at startup.
    client_thread = await ns_thread.clone(CLONE.NEWUSER|CLONE.NEWNET|CLONE.NEWNS|CLONE.NEWPID)
    # lightly sandbox by unmounting everything except for the executable and its deps (known via package manager)
    await unmount_everything_except(client_thread, miredo_exec.run_client.executable_path)
    # a helper function
    async def pass_fd(fd: FileDescriptor) -> str:
        await client_thread.inherit_fd(fd).disable_cloexec()
        return str(int(fd))
    client_child = await client_thread.execve(miredo_exec.run_client.executable_path, [
        miredo_exec.run_client.executable_path,
        await pass_fd(inet_sock), await pass_fd(tun_fd), await pass_fd(reqsock),
        await pass_fd(icmp6_fd), await pass_fd(privproc_pair.second),
        "teredo.remlab.net", "teredo.remlab.net"
    ])
    return Miredo(ns_thread, client_child, privproc_child)

async def start_miredo(nursery, miredo_exec: MiredoExecutables, thread: Thread) -> Miredo:
    miredo = await start_miredo_internal(thread, miredo_exec)
    nursery.start_soon(miredo.privproc_child.check)
    nursery.start_soon(miredo.client_child.check)
    return miredo

class TestMiredo(TrioTestCase):
    async def asyncSetUp(self) -> None:
        # TODO lmao stracing this stuff causes a bug,
        # what is even going on
        self.thread = local.thread
        print("a", time.time())
        self.exec = await MiredoExecutables.from_store(nix.local_store)
        print("a1", time.time())
        self.miredo = await start_miredo(self.nursery, self.exec, self.thread)
        print("b", time.time())
        self.netsock = await self.miredo.ns_thread.task.socket(AF.NETLINK, SOCK.DGRAM, NETLINK.ROUTE)
        print("b-1", time.time())
        print("b0", time.time())
        await self.netsock.bind(
            await self.miredo.ns_thread.ram.ptr(SockaddrNl(0, RTMGRP.IPV6_ROUTE)))
        print("b0.5", time.time())


    async def test_miredo(self) -> None:
        print("b1", time.time())
        ping6 = (await self.thread.environ.which("ping")).args('-6')
        print("b1.5", time.time())
        # TODO lol actually parse this, don't just read and throw it away
        await self.netsock.read(await self.miredo.ns_thread.ram.malloc(bytes, 4096))
        print("b2", time.time())
        thread = await self.miredo.ns_thread.clone()
        print("c", time.time())
        # bash = await rsc.which(self.thread, "bash")
        # await (await thread.exec(bash)).check()
        await add_to_ambient_caps(thread, {CAP.NET_RAW})
        await (await thread.exec(ping6.args('-c', '1', 'google.com'))).check()
        print("d", time.time())

if __name__ == "__main__":
    import unittest
    unittest.main()



"""Functions and classes for working with filesystem paths deployed with Nix

We use the nixdeps module to create build-time dependencies on Nix
derivations; setuptools will write out the paths and closures of the
Nix derivations we depend on.

We can use those dependencies by importing the `closure` module variables,
PackageClosure instances, which are independent of any specific thread.  To
turn a PackageClosure into a usable path, we can pass it to `deploy`,
which checks if the path is already deployed in that specific store,
and returns the path if so. If the path isn't deployed already,
`deploy` will deploy it for us.

"""
from __future__ import annotations
import typing as t
import os
import rsyscall.handle as handle
from rsyscall import local_thread
from rsyscall.thread import Thread, ChildThread
from rsyscall.command import Command
import trio
import struct
from dataclasses import dataclass
import nixdeps
from nixdeps import PackageClosure
import logging
from rsyscall.handle import WrittenPointer, Pointer, FileDescriptor
from rsyscall.path import Path

from rsyscall.sys.mount import MS
from rsyscall.fcntl import O
from rsyscall.sched import CLONE
from rsyscall.unistd import Pipe, OK

__all__ = [
    "copy_tree",
    "enter_nix_container",
    "deploy",
]

async def _exec_tar_copy_tree(src: ChildThread, src_paths: t.Sequence[t.Union[str, os.PathLike]], src_fd: FileDescriptor,
                              dest: ChildThread, dest_path: t.Union[str, os.PathLike], dest_fd: FileDescriptor) -> None:
    "Exec tar to copy files between two paths"
    dest_tar = await dest.environ.which("tar")
    src_tar = await dest.environ.which("tar")

    await dest.task.chdir(await dest.ram.ptr(dest_path))
    await dest.task.inherit_fd(dest_fd).dup2(dest.stdin)
    await dest_fd.close()
    dest_child = await dest.exec(dest_tar.args("--extract"))

    await src.task.inherit_fd(src_fd).dup2(src.stdout)
    await src_fd.close()
    src_child = await src.exec(src_tar.args(
        "--create", "--to-stdout", "--hard-dereference",
        "--owner=0", "--group=0", "--mode=u+rw,uga+r",
        *src_paths,
    ))
    await src_child.check()
    await dest_child.check()

async def copy_tree(src: Thread, src_paths: t.Sequence[t.Union[str, os.PathLike]], dest: Thread, dest_path: t.Union[str, os.PathLike]) -> None:
    """Copy all the listed `src_paths` to subdirectories of `dest_path`

    Example: if we pass src_paths=['/a/b', 'c'], dest_path='dest',
    then paths ['dest/b', 'dest/c'] will be created.
    """
    [(local_fd, dest_fd)] = await dest.connection.open_channels(1)
    src_fd = local_fd.move(src.task)
    await _exec_tar_copy_tree(await src.clone(), src_paths, src_fd,
                              await dest.clone(), dest_path, dest_fd)

async def _exec_nix_store_transfer_db(
        src: ChildThread, src_nix_store: Command, src_fd: FileDescriptor, closure: t.Sequence[t.Union[str, os.PathLike]],
        dest: ChildThread, dest_nix_store: Command, dest_fd: FileDescriptor,
) -> None:
    "Exec nix-store to copy the Nix database for a closure between two stores"
    await dest.task.inherit_fd(dest_fd).dup2(dest.stdin)
    await dest_fd.close()
    dest_child = await dest.exec(dest_nix_store.args("--load-db").env({'NIX_REMOTE': ''}))

    await src.task.inherit_fd(src_fd).dup2(src.stdout)
    await src_fd.close()
    src_child = await src.exec(src_nix_store.args("--dump-db", *closure))
    await src_child.check()
    await dest_child.check()

async def bootstrap_nix_database(
        src: Thread, src_nix_store: Command, closure: t.Sequence[t.Union[str, os.PathLike]],
        dest: Thread, dest_nix_store: Command,
) -> None:
    "Bootstrap the store used by `dest` with the necessary database entries for `closure`, coming from `src`'s store"
    [(local_fd, dest_fd)] = await dest.open_channels(1)
    src_fd = local_fd.move(src.task)
    await _exec_nix_store_transfer_db(await src.clone(), src_nix_store, src_fd, closure,
                                      await dest.clone(), dest_nix_store, dest_fd)

async def enter_nix_container(
        src: Thread, nix: PackageClosure,
        dest: Thread, dest_dir: Path,
) -> None:
    """Move `dest` into a container in `dest_dir`, deploying Nix inside.

    We can then use `deploy` to deploy other things into this container,
    which we can use from the `dest` thread or any of its children.

    """
    # we want to use our own container Nix store, not the global one on the system
    if 'NIX_REMOTE' in dest.environ:
        del dest.environ['NIX_REMOTE']
    # copy the binaries over
    await copy_tree(src, nix.closure, dest, dest_dir)
    # enter the container
    await dest.unshare(CLONE.NEWNS|CLONE.NEWUSER)
    await dest.mount(dest_dir/"nix", "/nix", "none", MS.BIND, "")
    # init the database
    nix_store = Command(nix.path/'bin/nix-store', ['nix-store'], {})
    await bootstrap_nix_database(src, nix_store, nix.closure, dest, nix_store)
    # add nix.path to PATH; TODO add a real API for this
    dest.environ.path.paths.append(Path(nix.path/'bin'))

async def deploy_nix_bin(src: Thread, nix: PackageClosure, dest: Thread) -> None:
    "Deploy the Nix binaries from `store` to /nix through `dest`"
    # copy the binaries over
    await copy_tree(src, nix.closure, dest, Path("/nix"))
    # init the database
    nix_store = Command(nix.path/'bin/nix-store', ['nix-store'], {})
    await bootstrap_nix_database(src, nix_store, nix.closure, dest, nix_store)

async def _exec_nix_store_import_export(
        src: ChildThread, src_nix_store: Command, src_fd: FileDescriptor,
        closure: t.Sequence[t.Union[str, os.PathLike]],
        dest: ChildThread, dest_nix_store: Command, dest_fd: FileDescriptor,
) -> None:
    "Exec nix-store to copy a closure of paths between two stores"
    await dest.task.inherit_fd(dest_fd).dup2(dest.stdin)
    await dest_fd.close()
    dest_child = await dest.exec(dest_nix_store.args("--import").env({'NIX_REMOTE': ''}))

    await src.task.inherit_fd(src_fd).dup2(src.stdout)
    await src_fd.close()
    src_child = await src.exec(src_nix_store.args("--export", *closure))
    await src_child.check()
    await dest_child.check()

async def _deploy(src: Thread, dest: Thread, path: PackageClosure) -> None:
    "Deploy a PackageClosure from the src Thread to the dest Thread"
    [(local_fd, dest_fd)] = await dest.open_channels(1)
    src_fd = local_fd.move(src.task)
    await _exec_nix_store_import_export(
        await src.clone(),
        await src.environ.which('nix-store'),
        src_fd, path.closure,
        await dest.clone(),
        await dest.environ.which('nix-store'),
        dest_fd)

class PackagePath(Path):
    "A Path with a few helper methods useful for Nix packages"
    def bin(self, name: str) -> Path:
        return Command(self/"bin"/name, [name], {})

async def deploy(thread: Thread, package: PackageClosure) -> PackagePath:
    "Deploy a PackageClosure to the filesystem of this Thread"
    if thread is not local_thread:
        # for remote threads, we need to check if it's actually there,
        # and deploy it if it's not there.
        # TODO we should really make and return a Nix GC root, too...
        ptr = await thread.ptr(package.path)
        try:
            await thread.task.access(ptr, OK.R)
        except (PermissionError, FileNotFoundError):
            await _deploy(local_thread, thread, package)
    return PackagePath(package.path)



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread

from rsyscall.unistd import Pipe
from rsyscall.fcntl import O
from rsyscall.sched import CLONE

class TestCat(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread
        self.cat = await self.thr.environ.which("cat")
        self.pipe_in = await (await self.thr.task.pipe(await self.thr.ram.malloc(Pipe))).read()
        self.pipe_out = await (await self.thr.task.pipe(await self.thr.ram.malloc(Pipe))).read()
        thread = await self.thr.clone()
        await thread.task.inherit_fd(self.pipe_in.read).dup2(thread.stdin)
        await thread.task.inherit_fd(self.pipe_out.write).dup2(thread.stdout)
        self.child = await thread.exec(self.cat)

    async def test_cat_pipe(self) -> None:
        in_data = await self.thr.ram.ptr(b"hello")
        written, _ = await self.pipe_in.write.write(in_data)
        valid, _ = await self.pipe_out.read.read(written)
        self.assertEqual(in_data.value, await valid.read())
        
        await self.pipe_in.write.close()
        await self.child.check()

    async def test_cat_async(self) -> None:
        stdin = await self.thr.make_afd(self.pipe_in.write, set_nonblock=True)
        stdout = await self.thr.make_afd(self.pipe_out.read, set_nonblock=True)
        in_data = await self.thr.ram.ptr(b"hello")
        written, _ = await stdin.write(in_data)
        valid, _ = await stdout.read(written)
        self.assertEqual(in_data.value, await valid.read())
        
        await self.pipe_in.write.close()
        await self.child.check()



import unittest
from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
from rsyscall.sys.socket import AF, SOCK, Socketpair
from rsyscall.unistd import Pipe, SEEK
from rsyscall.fcntl import O
from rsyscall.sys.wait import W, Siginfo

from rsyscall.sched import CLONE
from rsyscall.handle import Process
from rsyscall.tests.utils import assert_thread_works
from rsyscall.signal import SIG

class TestProc(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.local = local_thread
        self.init = await self.local.clone(CLONE.NEWUSER|CLONE.NEWPID)
        # set up proc

    async def test_pgid(self) -> None:
        try:
            last_pid = await self.init.task.open(await self.init.ptr("/proc/sys/kernel/ns_last_pid"), O.WRONLY)
        except FileNotFoundError:
            raise unittest.SkipTest("Requires /proc/sys/kernel/ns_last_pid, which requires CONFIG_CHECKPOINT_RESTORE")

        pgldr = await self.init.clone()
        await pgldr.task.setpgid()
        pgflr = await self.init.clone()
        await pgflr.task.setpgid(pgldr.process.process)
        self.assertEqual(int(await pgflr.task.getpgid()), 2)
        await pgldr.exit(0)
        await pgldr.process.waitpid(W.EXITED)
        self.assertIsNotNone(pgldr.process.process.death_state)
        if pgldr.process.process.death_state is None: raise Exception # for mypy
        self.assertEqual(pgldr.process.process.death_state.pid, 2)
        self.assertTrue(pgldr.process.process.death_state.died())
        self.assertEqual(int(await pgflr.task.getpgid()), 2)

        await self.init.spit(last_pid, b"1\n")

        with self.assertRaises(ProcessLookupError):
            await self.init.task._make_process(2).kill(SIG.NONE)
        pg_two = await self.init.clone()
        with self.assertRaises(ProcessLookupError):
            await self.init.task._make_process(2).kill(SIG.NONE)
        # Linux skips right over process 2, even though it's dead, because it's still used by the process group
        self.assertEqual(int(pg_two.task.process.near), 3)
